"""
Flatspin model
"""
#!/usr/bin/env python3
import sys
import os
import numpy as np
from numpy.linalg import norm
from math import ceil, sqrt
from scipy.spatial import cKDTree
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize, LinearSegmentedColormap
from matplotlib.collections import PatchCollection
import time
import itertools
from datetime import timedelta
from copy import copy
from PIL import Image
import pkg_resources
from abc import ABC, abstractmethod
import warnings

from flatspin.grid import Grid
from flatspin.vertices import find_vertices, vertex_type
from flatspin.utils import get_default_params
from flatspin.plotting import plot_vectors, plot_astroid
from flatspin.data import read_table

# TODO: Only import pycuda of use_cuda=1
normalize_rad = Normalize(vmin=0, vmax=2*np.pi)
normalize_vertex_type = Normalize(vmin=1, vmax=4)

kB = 1.38064852e-23 # m^2 kg s^-2 K^-1

class LabelIndexer:
    """ Lookup table of labels to spin index

    Supports standard numpy indexing such as slices etc.
    """
    def __init__(self, labels):
        labels = np.array(labels)
        if len(labels.shape) == 1:
            labels.shape += (1,)

        shape = tuple(np.max(labels, axis=0) + 1)
        # Empty slots in the lookup table are -1
        self.index = np.full(shape, -1, dtype=int)
        # Populate table with labels
        self.index[tuple(labels.T)] = np.arange(len(labels), dtype=int)

    def __getitem__(self, label):
        indices = self.index[label]
        indices = indices[indices >= 0]
        if len(indices) == 1:
            indices = indices[0]
        return indices

    def __repr__(self):
        return repr(self.index)

class SpinIce(ABC):
    """ Spin ice abstract base class

    The :class:`SpinIce` class contains the core implementation
    of the :ref:`flatspin model <model>`. It is an abstract base class, and
    cannot be instantiated directly. Subclasses of this class define the
    geometry, by implementing :func:`_init_geometry`.
    """
    def __init__(self, *, size=(4,4), lattice_spacing=1, hc=0.2, alpha=0.001,
            disorder=0, h_ext=(0,0), neighbor_distance=1, switching='sw',
            sw_b=0.41, sw_c=1, sw_beta=1.5, sw_gamma=3.9,
            temperature=0, therm_timescale=1, m_therm=.2*860e3*220e-9*80e-9*25e-9,
            attempt_freq=1e9, flip_mode='max', init='polarized',
            spin_axis=None, random_prob=0.5, random_seed=0,
            use_opencl=False, opencl_platform=0, opencl_device=0,
            use_cuda=False, astroid_resolution=1801):
        """
        Parameters
        ----------
        size : tuple of int
            The size of the spin ice (width, height).
            Note that the size is geometry specific.
        lattice_spacing : float
            Spacing between each spin.
        hc : float
            Mean switching threshold.
        alpha : float
            Coupling strength alpha = u_0 * M / (4 * pi * a^3), where a is the
            lattice spacing and M is the net magneticmoment of a single magnet.
        disorder : float
            Switching threshold disorder as percentage of hc. If non-zero,
            sample switching from a normal distribution with mean hc and
            standard deviation hc * disorder.
        h_ext : tuple of float
            External field (h_extx, h_exty).
        neighbor_distance : float
            Neighborhood to consider when calculating dipole interactions.
            All spins within lattice_spacing * neighbor_distance are considered
            neighbors. Set neighbor_distance=np.inf for a global neighborhood.
        switching : {'sw', 'budrikis'}
            Magnetic switching model:
                - 'sw': Extended Stoner-Wohlfarth model
                - 'budrikis': switching based on the parallel field component
        sw_b : float
            Height of the Stoner-Wohlfarth astroid (parallel axis).
        sw_c : float
            Width of the Stoner-Wohlfarth astroid (perpendicular axis).
        sw_beta : float
            Pointiness of the top/bottom of the Stoner-Wohlfarth astroid
            (parallel axis).
        sw_gamma : float
            Pointiness of the left/right of the Stoner-Wohlfarth astroid
            (perpendicular axis).
        temperature : float (positive)
            Absolute temperature in Kelvin.
        therm_timescale : float
            The thermal time scale (length in seconds of each simulated timestep).
            When temperature > 0, the thermal activity per timestep increases
            with therm_timescale. See also set_temperature().
        m_therm : float
            Thermal nucleation moment (thermal nucleation volume * magnetic moment).
            See also set_temperature().
        attempt_freq : float
            Attempt frequency (1/s).
            Only used for temperature calculations, see set_temperature().
        flip_mode : {'single', 'max', 'max-rand', 'all'}
            Flip mode strategy:
                - 'single': single uniform random flip (Budrikis Monte Carlo)
                - 'max': flip the spin with maximum energy, break ties by index
                - 'max-rand': flip the spin with maximum energy, break ties randomly
                - 'all': flip all flippable spins
        init : {1, -1, 'random', 'image.png'} or ndarray
            Initial spin state.
        spin_axis : float or None
            If not None, ensure positive spin directions are all along the
            given axis angle.
        use_opencl : bool
            Use OpenCL acceleration.
        opencl_platform : int
            Select OpenCL platform (if multiple GPUs of different families are installed).
        opencl_device : int
            Select OpenCL device (if multiple GPUs of the same family are installed).
        use_cuda : bool
            Use CUDA acceleration. NB! Requires compute platform 10 and above.
        astroid_resolution : number of polar points in one rotation to resolve
            astroid shape for h_dist calculcations

        """
        self.size = size
        self.lattice_spacing = lattice_spacing
        self.hc = hc
        self.alpha = alpha
        self.disorder = disorder
        self.switching = switching
        self.sw_params = (sw_b, sw_c, sw_beta, sw_gamma)
        self.m_therm = m_therm
        self.therm_timescale = therm_timescale
        self.attempt_freq = attempt_freq
        self.astroid_resolution = astroid_resolution
        self.flip_mode = flip_mode
        self.neighbor_distance = neighbor_distance
        self.random_prob = random_prob
        self.spin_axis = spin_axis
        assert switching in ('budrikis', 'sw')
        assert flip_mode in ('single', 'max', 'max-rand', 'all')

        self.labels = None

        self.cl = use_opencl
        self.cl_device = opencl_device
        self.cl_platform = opencl_platform
        self._cl_context = None

        self.cuda = use_cuda
        self._cuda_relax_kernel = None

        self.set_random_seed(random_seed)

        # geometry
        pos, angle = self._init_geometry()

        if self.spin_axis is not None:
            # ensure all angles are along spin axis
            a0 = np.deg2rad(self.spin_axis) - np.pi/2
            angle = (angle - a0) % np.pi + a0

        self.pos = pos.astype(float, order='C')
        self.angle = angle.astype(float)
        self.m = np.column_stack([np.cos(self.angle), np.sin(self.angle)])
        self.spin_count = len(self.pos)

        if self.labels is None:
            # No labels, labels = indices
            self.labels = np.arange(self.spin_count)

        self.L = LabelIndexer(self.labels)

        # astroid tree
        self._init_astroid_tree()

        # spin flip threshold
        self.threshold = np.zeros(self.spin_count)
        self._init_threshold()

        # spin state
        self.spin = np.ones(self.spin_count, dtype=np.int8)
        self._init_spin(init)

        # flat neighbor list (computed once)
        self._neighbor_list = self._init_neighbor_list()
        self.num_neighbors = self._neighbor_list.shape[-1]

        # dipolar fields between each spin and its neighbors
        # precomputed on demand by _h_dip/_init_cl
        self._h_dip_cache = None

        self.set_h_ext(h_ext)

        # minimum threshold from origo
        self._zero_to_astroid = self.astroid_tree.query([(0,0)], k=1)[0][0]
        self._multiflip_warn = 0.05

        # thermal noise
        self.temperature = 0
        self.set_temperature(temperature)
        self.update_thermal_noise()

    @property
    def N(self):
        """ Alias for self.spin_count """
        return self.spin_count

    def label(self, i):
        """ Get the label of a given spin index or list of indices """
        return tuple(self.labels[i])

    def indexof(self, label):
        """ Get the spin index of a given label or range of labels

        Alias for self.L[label]
        """
        return self.L[label]

    def indices(self):
        """ Get all spin indices """
        return range(self.spin_count)

    def all_indices(self):
        """ Get all spin indices as numpy integer array index
        See https://docs.scipy.org/doc/numpy/reference/arrays.indexing.html#integer-array-indexing
        """
        return list(self.indices())

    def _init_spin(self, init):
        if isinstance(init, str) and init == 'random':
            self.randomize(self.random_prob)

        if isinstance(init, str) and init.endswith('png'):
            self.set_spin_image(init)

        if isinstance(init, np.ndarray):
            self.set_spin(init)

        if isinstance(init, int) and init in (-1, 1):
            self.polarize(init)

    @abstractmethod
    def _init_geometry(self):
        """ Initialize the geometry

        Subclasses should override this method and return a tuple (pos, angle),
        where pos is an array with the (x, y) positions of the spins, while
        angle is an array with the rotations of the spins.

        The base class implementation may be invoked to generate an Ising
        geometry.
        """
        nx, ny = self.size
        spin_count = nx * ny
        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        # rotation of each spin: zero by default

        # positions of each spin
        rows = np.arange(ny, dtype=int)
        cols = np.arange(nx, dtype=int)
        grid = np.array(np.meshgrid(rows, cols))
        grid = grid.T.reshape((-1, 2))
        pos = self.lattice_spacing * np.flip(grid.astype(float), axis=-1)
        pos[:] = pos
        self.labels = grid

        return pos, angle

    def _init_threshold(self):
        # spin flip threshold
        if np.array(self.hc).size > 1:
           self._init_threshold_grid()
        else:
            if self.disorder:
                std = self.hc * self.disorder
                self.threshold = self.rng.normal(self.hc, std, self.threshold.shape)
            else:
                self.threshold[:] = self.hc

        if self.disorder:
            # all thresholds should be positive
            self.threshold = np.abs(self.threshold)

    def _init_threshold_grid(self):
        hc = np.atleast_2d(self.hc)
        G = self.fixed_grid((hc.shape[1], hc.shape[0]))

        if self.disorder:
            for i in np.ndindex(G.size):
                std = hc[i] * self.disorder
                spin_inds = G.point_index(i)
                self.threshold[spin_inds] = self.rng.normal(hc[i], std, len(spin_inds))
        else:
            spin_inds = self.all_indices()
            grid_inds = G.grid_index(spin_inds)
            self.threshold[spin_inds] = hc[grid_inds]

    def _init_neighbor_list(self):
        # Calculate neighborhood matrix
        # print(f'_init_neighbor_list(size={self.size})')
        neighbors = []
        num_neighbors = 0

        # Construct KDTree for every position
        tree = cKDTree(self.pos)

        nd = self.lattice_spacing * self.neighbor_distance
        nd += 1e-5 # pad to avoid rounding errors

        for i in self.indices():
            p = self.pos[i]
            n = tree.query_ball_point([p], nd)[0]
            n.remove(i)
            neighbors.append(n)
            num_neighbors = max(num_neighbors, len(n))

        # print(f'num_neighbors={num_neighbors}')

        # Neighborhood list, -1 marks end of each list
        neighbor_list = np.full((self.spin_count, num_neighbors), -1, dtype=np.int32)
        for i, neighs in enumerate(neighbors):
            neighbor_list[i,:len(neighs)] = neighs

        return neighbor_list

    def _init_h_dip(self):
        # when use_opencl=1, the hdip cache is initialized by _init_cl()
        assert (not self.cl) and (not self.cuda)
        assert self.num_neighbors > 0

        # cache dipolar fields for the neighborhood of all magnets
        _h_dip = np.zeros((self.spin_count, self.num_neighbors, 2), dtype=np.float64)
        for i in self.indices():
            for jj, j in enumerate(self.neighbors(i)):
                # print(f'i={i} jj={jj} j={j}')
                _h_dip[i][jj] = np.array(self.spin_dipolar_field(i, j))

        return _h_dip

    def _init_astroid_tree(self):
        # Switching astroid shape in units hk (self.threshold)
        b, c, beta, gamma = self.sw_params

        # Due to the two-fold symmetry of the astroid, we only need to consider
        # one quadrant where h_par <= 0 and h_perp >= 0
        # Note: the first element is top of astroid, while last element is the apex
        thetas = np.linspace(0, -np.pi/2, self.astroid_resolution)

        # h_par, h_perp are created from calculating h_par(h_perp) from SW-
        # curve, over thetas coresponding to negative h_pars
        h_perp = c * np.cos(thetas)
        h_par = -b * (1 - ((h_perp/c)**2)**(1/beta))**(gamma/2)

        # Construct cKDTree of astroid to quickly look up nearest point
        astroid_points = np.column_stack((h_par, h_perp))
        self.astroid_tree = cKDTree(astroid_points)

        # Calculate astroid normals based on left/right neighbor
        p0 = astroid_points[0:-2]
        p2 = astroid_points[2:]
        delta = p2 - p0
        dx = delta[:,0]
        dy = delta[:,1]
        normals = np.column_stack([dy, -dx])

        # Normalize to unit length
        lens = norm(normals, axis=1).reshape((-1, 1))
        normals /= lens

        # For edges of the astroid (top/bottom and apex), simply use the
        # normals of the next neighbor
        normals = np.pad(normals, ((1, 1), (0, 0)), 'edge')

        self.astroid_normals = normals

    def __getstate__(self):
        odict = copy(self.__dict__)
        # Don't pickle _cl attributes. They will be automatically
        # re-initialized by _init_cl() when needed
        for k in list(filter(lambda k: k.startswith('_cl'), odict)):
            odict[k] = None
        return odict

    def __eq__(self, other):
        state = self.__getstate__()
        other_state = other.__getstate__()
        for k in state:
            if np.any(state[k] != other_state[k]):
                return False
        return True

    @property
    def width(self):
        x = self.pos[:, 0]
        xmin, xmax = np.min(x), np.max(x)
        return xmax - xmin

    @property
    def height(self):
        y = self.pos[:, 1]
        ymin, ymax = np.min(y), np.max(y)
        return ymax - ymin

    @property
    def vectors(self):
        """ Spin vectors """
        spin = np.array(self.spin, dtype=float)
        spin.shape += (1,)
        mag = self.m * spin
        return mag

    def set_spin(self, spin):
        spin = np.array(spin, dtype=np.int8)
        assert spin.shape == (self.spin_count,)
        assert np.all(np.logical_or(spin == 1, spin == -1))
        self.spin = spin

    def set_spin_image(self, filename):
        """ Set spin state from image

        The image is resized to fit the default spin grid. Then the grayscale
        image is used to set spin state:
        * black maps to spin -1
        * non-black maps to spin 1
        """
        # Use default grid
        G = self.grid()
        # calculate size of grid
        height, width = G.size

        # load image
        img = Image.open(filename)
        img = img.convert('L') # convert to grayscale

        # flip it since (0,0) is top left in the image but (0,0) is bottom left
        # in our spin geometry
        img = img.transpose(Image.FLIP_TOP_BOTTOM)

        # resize image to the default grid size
        img = img.resize((width, height), resample=Image.Resampling.NEAREST)
        img = np.array(img)

        img[img != 0] = 1  # map non-black to spin 1
        img[img == 0] = -1 # map black to spin -1

        # finally set all the spins
        self.set_grid('spin', np.array(img))

    def set_threshold(self, threshold):
        """
        Set switching thresholds manually.

        NB: this method does not update self.hc
        """
        assert threshold.shape == (self.spin_count,)
        self.threshold = np.array(threshold)

    def set_hc(self, hc):
        """
        Set mean switching threshold hc.

        Rescales existing thresholds to have new mean hc and standard
        devication hc * disorder.
        """
        if np.array(hc).size > 1 or np.array(self.hc).size > 1:
            # old or new hc were grid based
            return self._set_hc_grid(hc)

        self.threshold *= hc / self.hc
        self.hc = hc

    def _set_hc_grid(self, hc):
        spin_inds = self.all_indices()

        # Determine the old hc values based on self.hc
        # which may or may not have been a grid based thing
        values = np.atleast_2d(self.hc)
        G = self.fixed_grid((values.shape[1], values.shape[0]))

        # Map old hc values onto spins
        old_hc = np.zeros(self.spin_count)
        grid_inds = G.grid_index(spin_inds)
        old_hc[spin_inds] = values[grid_inds]

        # Similarly, map new hc values onto spins
        values = np.atleast_2d(hc)
        G = self.fixed_grid((values.shape[1], values.shape[0]))
        grid_inds = G.grid_index(spin_inds)
        new_hc = np.zeros(self.spin_count)
        new_hc[spin_inds] = values[grid_inds]

        # Finally, rescale existing thresholds to have new mean hc from grid
        self.threshold *= new_hc / old_hc
        self.hc = hc

    def set_h_ext(self, h_ext):
        """
        Set external field to h_ext

        h_ext can either be a single vector for a uniform field (h_ext.shape==(2,))
        or a 2D array of vectors for a non-uniform field (h_ext.shape==(self.spin_count, 2))
        """
        h_ext = np.array(h_ext, dtype=np.float64, order='C')
        assert h_ext.shape == (2,) or h_ext.shape == (self.spin_count, 2)
        if h_ext.shape == (2,):
            h_ext = np.tile(h_ext, (self.spin_count, 1))
        self.h_ext = h_ext

    def set_h_ext_grid(self, h_ext):
        if h_ext.shape == (2,):
            # global
            self.set_h_ext(h_ext)
        else:
            # grid
            self.set_grid('h_ext', h_ext)

    def set_alpha(self, alpha):
        """ Set coupling strength alpha = u_0 * M / (4 * pi * a^3), where a is the
            lattice spacing and M is the net magneticmoment of a single magnet."""
        self.alpha = alpha

    def set_therm_timescale(self, therm_timescale):
        """ Set thermal time scale (length in seconds of each simulated timestep).
            When temperature > 0, the thermal activity per timestep increases
            with therm_timescale. See also set_temperature(). """
        self.therm_timescale = therm_timescale

    def set_m_therm(self, m_therm):
        """ Thermal nucleation moment (thermal nucleation volume * magnetic moment).
            See also set_temperature(). """
        self.m_therm = m_therm

    def set_temperature(self, temperature):
        """ Set temperature for the thermal field

        The random thermal field depends on temperature, as well as the
        parameters m_therm, therm_timescale and attempt_freq.

        Note: flatspin does not account for the temperature dependence of the
        parameters.  If these parameters are expected to vary significantly in
        the temperature range of interest, this has to be explicitly accounted
        for by the user.
        """
        assert self.m_therm > 0
        assert self.therm_timescale > 0
        assert self.attempt_freq > 0
        assert temperature >= 0

        # multiflip warning
        if self.temperature < temperature:
            # If the new temperature is higher, warn if multiflip is likely
            threshold_zero = np.mean(self.threshold) * self._zero_to_astroid
            f = self.attempt_freq * np.exp(-self.m_therm * \
                threshold_zero / (kB * temperature))
            P_flip_zero = np.exp(-f * self.therm_timescale)
            P_flip_once = f * self.therm_timescale * np.exp(-f * self.therm_timescale)
            P_multiflip = 1 - P_flip_zero - P_flip_once
            #print(f"P_flip_once={P_flip_once} P_multiflip={P_multiflip}")
            if P_multiflip > self._multiflip_warn:
                warnings.warn(f"Probability of multiflip, {P_multiflip:.3f}, "
                    + f"exceeds {self._multiflip_warn} for temperature={temperature}. "
                    + "Reduce therm_timescale or temperature.")

        self.temperature = temperature

    def set_random_seed(self, seed):
        """ Seed the random number generator used in the model. """
        self.random_seed = seed
        self.rng = np.random.default_rng(seed)

    def set_angle(self, angle):
        """ Change the angles of the spins.

        angle is an array of new angles in radians.
        The number of new angles must match the spin count.
        """
        angle = np.array(angle, dtype=float)
        assert angle.shape == (self.spin_count,)
        self.angle = angle
        self.m = np.column_stack([np.cos(angle), np.sin(angle)])

        # Invalidate the h_dip cache
        self._h_dip_cache = None
        if self.cl:
            self._init_cl_geometry()
        if self.cuda:
            self._init_cuda_geometry()

    def set_pos(self, pos):
        """ Change the positions of the spins.

        pos is an array of new positions.
        The number of new positions must match the spin count.
        """
        pos = np.array(pos, dtype=float, order='C')
        assert pos.shape == (self.spin_count, 2)
        self.pos = pos

        # Invalidate the h_dip cache
        self._h_dip_cache = None

        # Re-initialize neighbor list
        self._neighbor_list = self._init_neighbor_list()
        self.num_neighbors = self._neighbor_list.shape[-1]

        if self.cl:
            self._init_cl_geometry()
        if self.cuda:
            self._init_cuda_geometry()

    def set_geometry(self, pos, angle):
        """ Change the geometry.

        angle is an array of new angles in radians.
        pos is an array of new positions.
        The number of new positions and angles must match the spin count.
        """
        pos = np.array(pos, dtype=float, order='C')
        assert pos.shape == (self.spin_count, 2)
        angle = np.array(angle, dtype=float)
        assert angle.shape == (self.spin_count,)

        self.pos = pos
        self.angle = angle
        self.m = np.column_stack([np.cos(angle), np.sin(angle)])

        # Invalidate the h_dip cache
        self._h_dip_cache = None

        # Re-initialize neighbor list
        self._neighbor_list = self._init_neighbor_list()
        self.num_neighbors = self._neighbor_list.shape[-1]

        if self.cl:
            self._init_cl_geometry()
        if self.cuda:
            self._init_cuda_geometry()

    def set_neighbor_distance(self, neighbor_distance):
        """ Change the neighbor distance used to calculate dipolar fields """
        self.neighbor_distance = neighbor_distance

        # Invalidate the h_dip cache
        self._h_dip_cache = None

        # Re-initialize neighbor list
        self._neighbor_list = self._init_neighbor_list()
        self.num_neighbors = self._neighbor_list.shape[-1]

        if self.cl:
            self._init_cl_geometry()
        if self.cuda:
            self._init_cuda_geometry()

    def randomize(self, prob=0.5, seed=None):
        """ Randomize spins.

        Spins take the value -1 with probability prob.
        If seed is not None, re-seed the model RNG prior to randomizing spins.
        """
        if seed is not None:
            self.set_random_seed(seed)

        self.spin = -1 + 2 * self.rng.binomial(1, 1 - prob, self.spin_count)
        self.spin = self.spin.astype(np.int8)

    def polarize(self, spin=1):
        assert spin in (1, -1)
        self.spin[:] = spin

    def update_thermal_noise(self):
        """ Resamples thermal noise.
        Samples a field magnitude matching probability of switching from
        Poisson(arrhenius_rate(Zeeman/kB*T)*timestep) and returns field vector
        directed along minimal distance to astroid"""

        if self.temperature:
            draw_uniform = self.rng.uniform(size=self.spin_count)
            h_len = self._uniform_to_htherm(draw_uniform)
            self._h_therm_cache = self._field_astroid_direction(h_len)
            # _h_therm_cache is wrt spin direction (h_par, h_perp),
            # so convert to global reference frame to yield h_therm
            m = self.m * self.spin.reshape((-1, 1))
            m_x = np.column_stack([m[:,0], -m[:,1]])
            m_y = np.column_stack([m[:,1], m[:,0]])
            h_therm_x = np.sum(self._h_therm_cache * m_x, axis=1)
            h_therm_y = np.sum(self._h_therm_cache * m_y, axis=1)
            self.h_therm = np.column_stack([h_therm_x, h_therm_y])
        elif self.cuda:
            self._h_therm_cache = np.zeros((self.spin_count,2))
            self.h_therm = np.zeros((self.spin_count, 2))
        else:
            self._h_therm_cache = None
            self.h_therm = np.zeros((self.spin_count, 2))

    def neighbors(self, i):
        neighs = self._neighbor_list[i]
        return neighs[neighs >= 0]

    def spin_dipolar_field(self, i, j):
        """ Calculate dipolar field between spin i and j relative to positive spin """
        r = self.pos[j] - self.pos[i]
        m = self.m
        dist = norm(r)
        h_dip_1 = -m[j] / dist**3
        h_dip_2 = 3 * r * m[j].dot(r) / dist**5
        h_dip = h_dip_1 + h_dip_2

        m_angle = np.arctan2(m[i][1], m[i][0])
        h_angle = np.arctan2(h_dip[1], h_dip[0])
        h = norm(h_dip)
        theta = h_angle - m_angle
        h_par = h*np.cos(theta)
        h_perp = h*np.sin(theta)
        return np.array([h_par, h_perp], dtype=float)

    def dipolar_field(self, i):
        """ Calculate total dipolar field parallell to spin i """
        if self.cl:
            # path only for testing
            return self._h_dip_local_cl()[i]
        if self.cuda:
            return self._h_dip_local_cuda()[i]
        return self._h_dip_local(i)

    def dipolar_fields(self):
        if self.cl:
            return self._h_dip_local_cl()

        if self.cuda:
            return self._h_dip_local_cuda()

        h_dip = np.zeros((self.spin_count, 2))
        for i in self.indices():
            h_dip[i] = self.dipolar_field(i)
        return h_dip

    @property
    def _h_dip(self):
        if self._h_dip_cache is None:
            self._h_dip_cache = self._init_h_dip()
        return self._h_dip_cache

    def _h_dip_local(self, i):
        hdip = np.zeros(2) # h_par, h_perp
        for jj, j in enumerate(self.neighbors(i)):
            # print("h_dip", jj, j)
            hdip += self._h_dip[i][jj] * self.spin[i] * self.spin[j]
        return self.alpha * hdip

    def external_field(self, i):
        """ Calculate external field parallel and perpendicular to spin i """
        if self.cl:
            # path only for testing
            return tuple(self._external_fields_cl()[i])

        if self.cuda:
            return tuple(self._external_fields_cuda()[i])

        m = self.spin[i] * self.m[i]
        h_ext = self.h_ext[i]
        h_par = np.dot(h_ext, m)
        m_perp = [-m[1], m[0]]
        h_perp = np.dot(h_ext, m_perp)
        return np.array([h_par, h_perp], dtype=np.float64)

    def external_fields(self):
        """ Calculate external fields parallel and perpendicular to all spins """
        if self.cl:
            return self._external_fields_cl()

        if self.cuda:
            return self._external_fields_cuda()

        m = self.m * self.spin.reshape((-1, 1))
        h_par = np.sum(self.h_ext * m, axis=1)
        m_perp = np.column_stack([-m[:,1], m[:,0]])
        h_perp = np.sum(self.h_ext * m_perp, axis=1)
        return np.column_stack([h_par, h_perp])

    def thermal_field(self, i):
        """ Calculate thermal field parallel and perpendicular to spin i """
        if self._h_therm_cache is None:
            return np.zeros(2)
        # thermal field is already computerd in _h_therm_cache
        return (self._h_therm_cache[i])

    def thermal_fields(self):
        """ Calculate thermal fields parallel and perpendicular to all spins """
        if self._h_therm_cache is None:
            return np.zeros((self.spin_count, 2))
        # thermal field is already computerd in _h_therm_cache
        return self._h_therm_cache

    def total_field(self, i):
        """ Calculate the total field parallel to spin i """
        if self.cl:
            return self._total_fields_cl()[i] + self.thermal_field(i)

        if self.cuda:
            return self._total_fields_cuda()[i] + self.thermal_field(i)

        return self.dipolar_field(i) + self.external_field(i) + self.thermal_field(i)

    def total_fields(self):
        h_therm = self.thermal_fields() if self.temperature > 0 else 0
        if self.cl:
            # TODO: make thermal_fields_cl
            return self._total_fields_cl() + h_therm
        return self.dipolar_fields() + self.external_fields() + h_therm

    def total_non_thermal_fields(self):
        if self.cl:
            return self._total_fields_cl()
        return self.dipolar_fields() + self.external_fields()

    def flip(self, i):
        self.spin[i] *= -1
        if self._h_therm_cache is not None:
            # Flipping a spin effectively flips its thermal field
            self._h_therm_cache[i] *= -1

    # TODO: rename / discard?
    def _switching_energy_budrikis(self):
        h_tot = self.total_fields()
        h_par = h_tot[..., 0] # parallel components only
        E = -(h_par + self.threshold)
        return E

    def _switching_energy_sw(self):
        b, c, beta, gamma = self.sw_params

        h_tot = self.total_fields()
        h_par = h_tot[..., 0] / (b * self.threshold)
        h_perp = h_tot[..., 1] / (c * self.threshold)

        # Switching criteria 1: h_par**(2/3) + h_perp**(2/3) > hc**(2/3)
        # E > 0: switching is possible
        E = (h_par**2)**(1/gamma) + (h_perp**2)**(1/beta) - 1

        # Switching criteria 2: h_par < 0
        # make E negative when h_par >= 0
        # Use a very small number here instead of zero to avoid a corner case:
        # when the field is perfectly perpendicular to the magnet, rounding
        # errors can cause h_par to always be negative regardless of the spin,
        # resulting in an infinite switching loop.
        cond = h_par >= -1e-12
        E[cond] = -np.abs(E[cond])
        return E

    def switching_energy(self):
        if self.switching == 'budrikis':
            E = self._switching_energy_budrikis()
        else:
            E = self._switching_energy_sw()

        return E

    def flippable_energy(self):
        flippable = []
        energy = self.switching_energy()
        flippable = np.nonzero(energy > 0)
        energy = energy[flippable]
        flippable = flippable[0]
        return flippable, energy

    def flippable(self):
        return self.flippable_energy()[0]

    def _field_astroid_direction(self, magnitude):
        """ Return a vector in the direction of shortest path to astroid """
        if self.switching == 'budrikis':
            return self._field_astroid_direction_budrikis(magnitude)

        return self._field_astroid_direction_sw(magnitude)

    def _field_astroid_direction_budrikis(self, magnitude):
        """ Return a vector in the direction of shortest path to astroid """
        ones = np.ones(self.spin_count)
        h_therm = np.column_stack([-magnitude * ones, 0 * ones])
        return h_therm

    def _field_astroid_direction_sw(self, magnitude):
        """ Return a vector in the direction of shortest path to astroid """
        b, c, beta, gamma = self.sw_params

        # Calculate h_par / h_perp
        h_tot = self.total_non_thermal_fields()

        # Normalize to astroid units
        h_tot /= self.threshold.reshape((-1, 1))

        # Astroid is symmetric, and we only sample the quadrant where h_perp is
        # positive.  When h_perp is negative, we reflect around the h_par axis,
        # and flip the corresponding directions back later.
        h_perp_negative = h_tot[:,1] < 0
        h_tot[:,1] = np.abs(h_tot[:,1])

        h_par = h_tot[..., 0]
        h_perp = h_tot[..., 1]

        # Magnitude is also normalized to astroid units
        if np.isscalar(magnitude):
            magnitude = magnitude * np.ones(self.spin_count)
        else:
            magnitude = np.array(magnitude)
        magnitude /= self.threshold

        # Look up nearest point on astroid
        astroid_points = self.astroid_tree.data
        distances, indices = self.astroid_tree.query(h_tot, k=1)
        nearest = astroid_points[indices]

        # The normal of the nearest point is the direction towards the shortest
        # path out of the astroid... except for a few corner cases, which are
        # handled below.
        normals = self.astroid_normals[indices]

        #
        # When the astroid is not smooth, we cannot rely on the normal.
        # For these cases, we calculate the direction manually.
        #

        # When beta > 2, the astroid is not smooth at the apex
        if beta > 2:
            # If the nearest point is the apex...
            apex_index = len(self.astroid_tree.data) - 1
            cond = (indices == apex_index)

            # Only consider the case where we are on the left side of the apex.
            # The approximation is good enough close to the apex, and this
            # avoids the possibility of crossing the astroid twice when
            # calculating the direction manually, below.  If we didn't ignore
            # this case, it's possible to cross the astroid twice without also
            # crossing the h_perp=0 axis, which would circumvent our
            # double-crossing avoidance code (later).
            cond[cond] = h_par[cond] < nearest[cond,0]

            # Calculate the direction manually based on the next nearest point.
            # We use the next nearest point instead of the apex to ensure the
            # perpendicular component is never zero, which would cause trouble
            # in the double-crossing avoidance code (later).
            next_point = astroid_points[apex_index - 1]
            new_normals = h_tot[cond] - next_point
            lens = norm(new_normals, axis=1).reshape((-1, 1))
            new_normals /= lens
            normals[cond] = new_normals

        # The top/bottom of the astroid is not smooth: it is effectively a
        # corner between the astroid and the h_par=0 axis.

        # If the nearest point is at the top (or bottom)...
        index_top = 0
        cond = indices == index_top
        # ... and h_par is positive (we're inside the astroid, in all other
        # cases the direction based on the normal is fine)
        cond[cond] = h_par[cond] > 0

        # Calculate the direction manually towards the top astroid point
        new_normals = astroid_points[index_top] - h_tot[cond]
        lens = norm(new_normals, axis=1).reshape((-1, 1))
        new_normals /= lens
        normals[cond] = new_normals

        #
        # End of special cases where the astroid is non-smooth
        #

        # If we are above/below the astroid, the shortest distance might be
        # horizontally until we cross the h_par = 0 axis
        cond = h_perp > c
        cond[cond] = distances[cond] > np.abs(h_par[cond])
        normals[cond] = [-1, 0]

        # Direction out of astroid is simply the normals scaled with the magnitudes
        h_dir = normals * magnitude.reshape((-1, 1))

        # Ensure we cannot cross astroid boundary more than once.
        # Negative magnitudes represent thermal energy which counteracts
        # field-based switching, i.e., starting outside the astroid, a thermal
        # field can enter the astroid again.
        # If the magnitude is negative and sufficiently strong, a thermal field may
        # first bring us into the astroid (correct), but then out again (incorrect).
        # Prevent this from happening by calculating the distance to the
        # h_perp=0 axis and make sure it is never crossed.
        #cond = magnitude < 0 and np.abs(h_dir[:,1]) > 1e-5

        # Where do we end up after the thermal field is applied?
        dest = h_perp + h_dir[:,1]
        # Check whether we cross the h_perp=0 axis
        # Due to symmetry, we always start above the axis
        #cond = np.logical_and(np.sign(h_perp) != 0, np.sign(h_perp) != np.sign(dest))
        cond = np.sign(dest) < 0
        # ... and that magnitude is negative
        cond = np.logical_and(cond, magnitude < 0)
        # Reduce the magnitude of h_dir until we hit the h_perp=0 axis
        scale = -h_perp[cond] / h_dir[cond,1]
        h_dir[cond] *= scale.reshape((-1, 1))

        # Flip any vectors with negative h_perp
        h_dir[h_perp_negative,1] *= -1

        # Finally, convert back to h_ext units
        h_dir *= self.threshold.reshape((-1, 1))

        return h_dir

    def _uniform_to_htherm(self, u):
        ''' Transform a random uniform variable to a thermal field magnitude
        '''

        T = self.temperature
        m_therm = self.m_therm
        attempt_freq = self.attempt_freq
        dt = self.therm_timescale

        return (-kB*T / m_therm) * np.log(np.log(u) / (-dt * attempt_freq))

    def step(self):
        """ Perform a flip of one or more flippable spins """
        flip, energy = self.flippable_energy()
        # print("flippable:", len(flip))
        if len(flip) == 0:
            return False

        # TODO: rename / parameterize?
        if self.flip_mode == 'single':
            # Flip a single random spin
            i = self.rng.choice(flip)
            self.flip(i)

        elif self.flip_mode == 'max':
            # Flip the spin with maximum energy
            i = flip[np.argmax(energy)]
            self.flip(i)

        elif self.flip_mode == 'max-rand':
            # Flip the spin with maximum energy, ties broken by random choice
            # Break ties by random choice
            i = np.argmax(energy)
            ties, = np.nonzero(energy == energy[i])
            j = flip[self.rng.choice(ties)]
            self.flip(j)

        # TODO: drop?
        elif self.flip_mode == 'all':
            for i in flip:
                self.flip(i)

        return True

    def relax(self):
        """ Flip spins until equilibrium, return number of calls to step() """
        return self._relax(True, True)

    def _relax(self, copy_cpu_to_gpu=True, copy_gpu_to_cpu=True):
        """ Optimized relax() to avoid unnecessary copying """
        self.update_thermal_noise()

        if self.cuda:
            return self._relax_cuda(copy_cpu_to_gpu, copy_gpu_to_cpu)

        steps = 0

        while self.step():
            steps += 1

        return steps

    def _relax_cuda(self, copy_cpu_to_gpu=True, copy_gpu_to_cpu=True):
        """ Flip spins until equilibrium, return number of calls to step() """
        if not self._cuda_relax_kernel:
            self._init_cuda()

        #hack to avoid copying data if we're not using thermal fields
        if self._h_therm_cache is not None:
            cuda.memcpy_htod(self._cuda_tmp_temp_field, self._h_therm_cache)

        if copy_cpu_to_gpu:
            cuda.memcpy_htod(self._cuda_spin, self.spin)

        cuda.memcpy_htod(self._cuda_h_ext, self.h_ext)

        b, c, beta, gamma = self.sw_params

        self._cuda_relax_kernel(
                (1,1),
                (1,1,1),
                self._cuda_h_dip_cache,
                self._cuda_m,
                self._cuda_h_ext,
                self._cuda_tmp_dip_field,
                self._cuda_tmp_ext_field,
                self._cuda_tmp_temp_field,
                self._cuda_threshold,
                self._cuda_tmp_switching_energies,
                self._cuda_spin,
                self._cuda_neighbor_list,
                self._cuda_num_steps,
                self.alpha,
                b, c, beta, gamma,
                self.num_neighbors,
                self.spin_count,
                self._cuda_done)

        #cuda.memcpy_dtoh(self.spin, self._cuda_spin)
        if (copy_gpu_to_cpu):
            cuda.memcpy_dtoh(self.spin, self._cuda_spin)

        _cuda_returned_num_steps = np.array([0], dtype=np.int32)
        cuda.memcpy_dtoh(_cuda_returned_num_steps, self._cuda_num_steps)

        return _cuda_returned_num_steps[0]

    def energy(self):
        """ Calculate energy per spin

        The energy is the component of the total field directed
        anti-parallel to the spin magnetization """

        h_tot = self.total_fields()
        h_par = h_tot[..., 0] # parallel components only
        E = -h_par
        return E

    def dipolar_energy(self):
        """ Calculate the dipolar energy per spin

        The dipolar energy is the component of the dipolar field directed
        anti-parallel to the spin magnetization """
        h_dip = self.dipolar_fields()
        h_par = h_dip[..., 0] # parallel components only
        E = -h_par
        return E

    def external_energy(self):
        """ Calculate the external energy per spin

        The external energy is the component of the external field directed
        anti-parallel to the spin magnetization """
        h_ext = self.external_fields()
        h_par = h_ext[..., 0] # parallel components only
        E = -h_par
        return E

    def thermal_energy(self):
        """ Calculate the thermal energy per spin

        The thermal energy is the component of the thermal field directed
        anti-parallel to the spin magnetization """
        h_therm = self.thermal_fields()
        h_par = h_therm[..., 0] # parallel components only
        E = -h_par
        return E

    def total_energy(self):
        """ Calculate the total energy """
        return np.sum(self.energy())

    def total_dipolar_energy(self):
        """ Calculate the total dipolar energy """
        return np.sum(self.dipolar_energy())

    def total_external_energy(self):
        """ Calculate the total external energy """
        return np.sum(self.external_energy())

    def total_thermal_energy(self):
        """ Calculate the total thermal energy """
        return np.sum(self.thermal_energy())

    def total_energies(self):
        dip = self.total_dipolar_energy()
        ext = self.total_external_energy()
        therm = self.total_thermal_energy()
        total = np.sum((dip, therm, ext))
        return [dip, ext, therm, total]

    def total_magnetization(self):
        mag = self.vectors
        return np.array(np.sum(mag, axis=0))

    # sub-classes can override this to specify the size of the vertex window
    # (see find_vertices)
    _vertex_size = (2, 1)

    def find_vertices(self):
        """ Find the vertices in this geometry

        Returns a tuple (vi, vj, indices) where vi, vj are the vertex indices
        and indices is a list of spin indices corresponding to each vertex
        index.
        """
        return find_vertices(self.grid(), self.pos, self.angle, self._vertex_size)

    def vertices(self):
        """ Get the spin indices of all vertices """
        vi, vj, indices = self.find_vertices()
        return indices

    def vertex_indices(self):
        """ Get all vertex indices """
        vi, vj, indices = self.find_vertices()
        return vi, vj

    def vertex_type(self, v):
        """ Get the vertex type for a vertex where v are the spin indices
        of the vertex """
        return vertex_type(self.spin[v], self.pos[v], self.angle[v])

    def vertex_count(self):
        """ Count the number of vertices of each type

        Returns a tuple (types, counts) where types are the different vertex
        types and counts are the corresponding counts. """
        vertex_types = [self.vertex_type(v) for v in self.vertices()]
        types, counts = np.unique(vertex_types, return_counts=True)
        return tuple(types), tuple(counts)

    def vertex_population(self):
        """ Calculate the vertex type population as a fraction of all vertices

        Returns a tuple (types, pops) where types are the different vertex
        types and pops are the corresponding fractions. """
        types, counts = self.vertex_count()
        counts = np.array(counts)
        pops = counts / np.sum(counts)
        return types, tuple(pops)

    def vertex_pos(self, v):
        """ Get the position of a vertex where v are the spin indices of the
        vertex """
        return np.mean(self.pos[v], axis=0)

    def vertex_mag(self, v):
        """ Get the direction of a vertex v """
        spin = self.spin[v]
        spin.shape += (1,)
        m = self.m[v]
        mag = m * spin
        return np.sum(mag, axis=0)

    @property
    def _default_cell_size(self):
        return (self.lattice_spacing, self.lattice_spacing)

    def grid(self, cell_size=None):
        """ Map spin indices onto a regular grid

        The spacing between each grid point is given by cell_size.
        If cell_size <= lattice_spacing, each grid cell will contain at
        most one spin.
        If cell_size > lattice_spacing, each grid cell may contain more
        than one spin.
        If cell_size is None, an optimal (geometry dependent) cell size is used

        Returns a Grid object which allows quick lookup of spin index to
        grid index
        """
        if cell_size is None:
            cell_size = self._default_cell_size
        if np.isscalar(cell_size):
            cell_size = (cell_size, cell_size)

        # Calculate optimal padding
        padx = min(self._default_cell_size[0]/2, cell_size[0]/2)
        pady = min(self._default_cell_size[1]/2, cell_size[1]/2)
        padding = (padx, pady)

        grid = Grid(self.pos, cell_size, padding)

        return grid

    def fixed_grid(self, grid_size):
        """ Map spin indices onto a regular grid of fixed size

        Like grid() but takes grid size (number of cells) as parameter
        instead of cell size.
        """
        return Grid.fixed_grid(self.pos, grid_size)

    def set_grid(self, attr, values):
        """ Map grid values onto some spin attribute.

        Valid attributes: spin, h_ext, threshold
        """
        values = np.array(values)
        G = self.fixed_grid((values.shape[1], values.shape[0]))

        # spin indices as numpy integer array index
        spin_inds = self.all_indices()

        # map all spin indices to grid index
        grid_inds = G.grid_index(spin_inds)

        # determine shape of attr and the corresponding set-function
        set_attr = getattr(self, 'set_' + attr)
        oldattr = getattr(self, attr)
        newattr = oldattr.copy()
        newattr[spin_inds] = values[grid_inds]
        set_attr(newattr)

    def view_grid(self, attr, cell_size=None, method='sum'):
        """ Project some spin attribute onto a grid.

        Valid attributes: spin, vectors, h_ext, threshold, pos

        The spacing between each grid cell is given by cell_size.
        If cell_size <= lattice_spacing, each grid cell will contain at
        most one spin attribute.
        If cell_size > lattice_spacing, each grid cell will contain the
        sum of several spin attributes.
        """

        # make the grid
        G = self.grid(cell_size)

        # get the values we wish to view on the grid
        values = getattr(self, attr)

        return G.add_values(values, method=method)

    # TODO: return spin instead of vectors
    def spin_grid(self, cell_size=None):
        """ Project the spin vectors onto a grid.

        See view_grid() for information about cell_size
        """
        return self.view_grid('vectors', cell_size)

    def plot(self, style="arrow", **kwargs):
        return plot_vectors(self.pos, self.vectors, style=style, **kwargs)

    def plot_energy(self, **kwargs):
        E = self.energy()
        kwargs.setdefault('cmap', 'bwr')
        return self.plot(C=E, **kwargs)

    def plot_vertices(self, **kwargs):
        size = self.lattice_spacing * 0.15
        vertices = self.vertices()
        circles = [plt.Circle(self.vertex_pos(v), size) for v in vertices]
        colors = np.array([self.vertex_type(v) for v in vertices])
        col = PatchCollection(circles, cmap='vertex-type', norm=normalize_vertex_type, **kwargs)
        col.set_array(colors)
        ax = plt.gca()
        ax.add_collection(col)
        ax.autoscale_view()
        return col

    def plot_vertex_mag(self, style="arrow", **kwargs):
        vertices = self.vertices()
        XY = np.array([self.vertex_pos(v) for v in vertices])
        UV = np.array([self.vertex_mag(v) for v in vertices])

        # Normalize vectors to unit length
        nmax = norm(UV, axis=-1).max()
        if nmax != 0:
            UV = UV / nmax

        return plot_vectors(XY, UV, style=style, **kwargs)

    def plot_astroid(self, rotation=0, **kwargs):
        b, c, beta, gamma = self.sw_params
        return plot_astroid(b, c, beta, gamma, rotation=rotation, **kwargs)

    def plot_astroids(self, **kwargs):
        angles = np.unique(self.angle)
        b, c, beta, gamma = self.sw_params
        for a in angles:
            rotation = np.rad2deg(a)
            kwargs['label'] = f"{rotation:.1f}°"
            plot_astroid(b, c, beta, gamma, rotation=rotation, **kwargs)
        plt.legend()

    def _init_h_dip_cl(self):
        # Only called by _init_cl()
        assert self.cl
        assert self.num_neighbors > 0

        posr = self.pos.ravel()
        pos_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=posr)
        res_g = cl.Buffer(self._cl_context, clmf.WRITE_ONLY, 2 * self.spin_count * self.num_neighbors * np.dtype(np.float64).itemsize)
        self._cl_prg.spin_dipolar_field(
                self._cl_queue,
                (self.spin_count, self.num_neighbors),
                None,
                pos_g,
                res_g,
                self._cl_neighbors_g,
                self._cl_n_neighbors_g,
                self._cl_m_g
                )


        res = np.empty((self.spin_count, self.num_neighbors, 2), dtype=np.float64)

        cl.enqueue_copy(self._cl_queue, res, res_g)

        return res

    def _init_h_dip_cuda(self):
        # Only called by _init_cuda()
        assert self.cuda
        assert self.num_neighbors > 0

        self._cuda_spin_dipolar_field = self.mod.get_function('spin_dipolar_field').prepare("PPPiP").prepared_call

        nn = max(1, self.num_neighbors)
        posr = self.pos.ravel()
        _cuda_pos_g = cuda.mem_alloc(posr.nbytes)
        _cuda_res_g = cuda.mem_alloc(2 * self.spin_count * nn * np.dtype(np.float64).itemsize)

        cuda.memcpy_htod(_cuda_pos_g, posr)
        #cuda.memcpy_htod(_cuda_res_g,

        #blocksize = 1024;
        #grid_x_dim = np.ceil(self.spin_count/blocksize);
        #grid_y_dim = np.ceil(self.num_neighbors/blocksize);
        grid_x_dim = self.spin_count
        grid_y_dim = nn

        self._cuda_spin_dipolar_field(
                (grid_x_dim, grid_y_dim),
                (1,1,1),
                _cuda_pos_g,
                _cuda_res_g,
                self._cuda_neighbor_list,
                self.num_neighbors,
                self._cuda_m
                )

        res = np.zeros((self.spin_count, nn, 2), dtype=np.float64)

        cuda.memcpy_dtoh(res, _cuda_res_g)

        return res



    def _init_cl(self):
        assert self.cl
        if self._cl_context:
            # already initialized
            return

        global cl, clmf
        cl = __import__("pyopencl")
        clmf = cl.mem_flags

        platforms = cl.get_platforms()
        assert(len(platforms) > 0), "Couldn't find any platforms! Driver installed?"
        platform = platforms[self.cl_platform]

        devices = platform.get_devices()
        assert(len(devices) > 0), "No devices in platform"
        device = devices[self.cl_device]
        # print("Using device {} on {}".format(device.name, platform.name))

        self._cl_context = cl.Context([device])
        self._cl_queue = cl.CommandQueue(self._cl_context)

        src = pkg_resources.resource_string(__name__, "flatspin_cl_kernels.c")
        src = src.decode("utf-8")

        self._cl_prg = cl.Program(self._cl_context, src).build(options=[])

        #preload some of the buffers.
        self._cl_spin_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self.spin)
        self._cl_res_g = cl.Buffer(self._cl_context, clmf.WRITE_ONLY,
                np.prod((self.spin_count, 2)) * np.dtype(np.float64).itemsize)
        self._cl_alpha_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=np.float64(self.alpha))
        self._cl_h_ext_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self.h_ext)

        self._init_cl_geometry()

    def _init_cl_geometry(self):
        """ (Re-)initialize cl buffers that depend on geometry """
        global cl, clmf
        assert self.cl
        if not self._cl_context:
            # CL not yet initialized, nothing to do
            return

        self._cl_n_neighbors_g = cl.Buffer(self._cl_context, clmf.READ_ONLY |
                clmf.COPY_HOST_PTR, hostbuf=np.int32(self.num_neighbors))
        self._cl_m_g = cl.Buffer(self._cl_context, clmf.READ_ONLY |
                clmf.COPY_HOST_PTR, hostbuf=np.float64(self.m))

        # These buffers are NULL when there are zero neighbors
        self._cl_neighbors_g = None
        self._cl_hdip_cache_g = None

        if self.num_neighbors > 0:
            self._cl_neighbors_g = cl.Buffer(self._cl_context, clmf.READ_ONLY |
                    clmf.COPY_HOST_PTR, hostbuf=self._neighbor_list)
            # we are now ready to calculate h_dip cache on the GPU
            if self._h_dip_cache is None:
                self._h_dip_cache = self._init_h_dip_cl()
                self._cl_hdip_cache_g = cl.Buffer(self._cl_context,
                        clmf.READ_ONLY | clmf.COPY_HOST_PTR,
                        hostbuf=self._h_dip_cache)




    def _init_cuda(self):
        assert self.cuda
        if self._cuda_relax_kernel:
            return

        import pycuda.autoinit
        import pycuda
        from pycuda.compiler import DynamicSourceModule
        global cuda
        cuda = pycuda.driver

        # print("Initializing CUDA")
        src = pkg_resources.resource_string(__name__, "flatspin_cuda_kernels.cu")
        src = src.decode("utf-8")
        self.mod = DynamicSourceModule(src)
        self._cuda_relax_kernel = self.mod.get_function('relax').prepare("PPPPPPPPPPPdddddiiP").prepared_call

        self._cuda_spin = cuda.mem_alloc(self.spin_count * np.dtype('int8').itemsize)
        cuda.memcpy_htod(self._cuda_spin, self.spin)

        self._cuda_tmp_dip_field = cuda.mem_alloc(int(2 * self.spin_count * np.dtype('double').itemsize))
        self._cuda_tmp_ext_field = cuda.mem_alloc(int(2 * self.spin_count * np.dtype('double').itemsize))
        self._cuda_tmp_temp_field = cuda.mem_alloc(int(2 * self.spin_count * np.dtype('double').itemsize))

        #transfer 0's to the card
        cuda.memcpy_htod(self._cuda_tmp_temp_field, np.zeros((self.spin_count, 2), dtype=np.float64))

        self._cuda_h_ext = cuda.mem_alloc(self.h_ext.nbytes)
        cuda.memcpy_htod(self._cuda_h_ext, self.h_ext)

        self._cuda_threshold = cuda.mem_alloc(self.threshold.nbytes)
        cuda.memcpy_htod(self._cuda_threshold, self.threshold)

        self._cuda_num_steps = cuda.mem_alloc(4) #np.array([0], dtype=np.int32)

        self._cuda_tmp_switching_energies = cuda.mem_alloc(self.spin_count * np.dtype('double').itemsize)

        self._cuda_done = cuda.mem_alloc(np.dtype('int').itemsize)

        self._init_cuda_geometry()

    def _init_cuda_geometry(self):
        """ (Re-)initialize cuda buffers that depend on geometry """
        global cuda
        assert self.cuda
        if not self._cuda_relax_kernel:
            # CUDA not yet initialized, nothing to do
            return

        self._cuda_m = cuda.mem_alloc(self.m.nbytes)
        cuda.memcpy_htod(self._cuda_m, self.m)

        # These buffers are NULL when there are zero neighbors
        self._cuda_neighbor_list = 0
        self._cuda_h_dip_cache = 0

        if self.num_neighbors > 0:
            self._cuda_neighbor_list = cuda.mem_alloc(self._neighbor_list.nbytes)
            cuda.memcpy_htod(self._cuda_neighbor_list, self._neighbor_list)

            if self._h_dip_cache is None:
                self._h_dip_cache = self._init_h_dip_cuda()

            self._cuda_h_dip_cache = cuda.mem_alloc(self._h_dip_cache.nbytes)
            cuda.memcpy_htod(self._cuda_h_dip_cache, self._h_dip_cache)

    def _total_fields_cl(self):
        self._init_cl()

        #this will calcualte for all indices
        #copy the updated data to the device
        #print(self.spin.shape)

        cl.enqueue_copy(self._cl_queue, self._cl_spin_g, self.spin)
        cl.enqueue_copy(self._cl_queue, self._cl_h_ext_g, self.h_ext)
        cl.enqueue_copy(self._cl_queue, self._cl_alpha_g, np.float64(self.alpha))

        self._cl_prg.total_fields(self._cl_queue, (self.spin_count,), None,
                self._cl_spin_g,
                self._cl_hdip_cache_g,
                self._cl_alpha_g,
                self._cl_res_g,
                self._cl_neighbors_g,
                self._cl_n_neighbors_g,
                self._cl_m_g,
                self._cl_h_ext_g
        )

        res = np.empty((self.spin_count, 2), dtype=np.float64)
        cl.enqueue_copy(self._cl_queue, res, self._cl_res_g)

        return res

    #Eh, it's just the two calls isolated anyway.
    def _total_fields_cuda(self):
        self._init_cuda()
        return self._h_dip_local_cuda() + self._external_fields_cuda()

    def _h_dip_local_cl(self):
        self._init_cl()

        #this will calcualte for all indices
        cl.enqueue_copy(self._cl_queue, self._cl_spin_g, self.spin)
        cl.enqueue_copy(self._cl_queue, self._cl_alpha_g, np.float64(self.alpha))


        self._cl_prg.test_h_dip_local(self._cl_queue, (self.spin_count,), None,
                self._cl_spin_g,
                self._cl_hdip_cache_g,
                self._cl_alpha_g,
                self._cl_res_g,
                self._cl_neighbors_g,
                self._cl_n_neighbors_g
        )

        res = np.empty((self.spin_count, 2), dtype=np.float64)
        cl.enqueue_copy(self._cl_queue, res, self._cl_res_g)

        return res

    def _external_fields_cl(self):
        self._init_cl()

        #this will calcualte for all indices
        spin_g = cl.Buffer(self._cl_context, clmf.READ_ONLY | clmf.COPY_HOST_PTR, hostbuf=self.spin)

        cl.enqueue_copy(self._cl_queue, self._cl_spin_g, self.spin)
        cl.enqueue_copy(self._cl_queue, self._cl_h_ext_g, self.h_ext)


        self._cl_prg.test_h_ext(self._cl_queue, (self.spin_count,), None,
                self._cl_spin_g,
                self._cl_m_g,
                self._cl_h_ext_g,
                self._cl_res_g,
        )

        res = np.empty((self.spin_count, 2), dtype=np.float64)
        cl.enqueue_copy(self._cl_queue, res, self._cl_res_g)

        return res

    def _external_fields_cuda(self):
        self._init_cuda()
        external_field_kernel = self.mod.get_function('external_field').prepare("PPPPi").prepared_call
        cuda.memcpy_htod(self._cuda_spin, self.spin)
        cuda.memcpy_htod(self._cuda_h_ext, self.h_ext)
        external_field_kernel(
                                (self.spin_count, 1),
                                (1,1,1),
                                self._cuda_spin,
                                self._cuda_m,
                                self._cuda_h_ext,
                                self._cuda_tmp_ext_field,
                                self.spin_count)

        res = np.zeros((self.spin_count, 2), dtype=np.float64)
        cuda.memcpy_dtoh(res, self._cuda_tmp_ext_field)
        return res


    def _h_dip_local_cuda(self):
        self._init_cuda()

        cuda.memcpy_htod(self._cuda_spin, self.spin)
        cuda.memcpy_htod(self._cuda_h_ext, self.h_ext)

        dipolar_field_kernel = self.mod.get_function('h_dip_local').prepare("PdPPiPi").prepared_call

        dipolar_field_kernel(
                                (self.spin_count, 1),
                                (1,1,1),
                                self._cuda_h_dip_cache,
                                self.alpha,
                                self._cuda_spin,
                                self._cuda_neighbor_list,
                                self.num_neighbors,
                                self._cuda_tmp_dip_field,
                                self.spin_count)

        res = np.zeros((self.spin_count, 2), dtype=np.float64)
        cuda.memcpy_dtoh(res, self._cuda_tmp_dip_field)

        return res


class IsingSpinIce(SpinIce):
    def __init__(self, *, spin_angle=90, **kwargs):
        self.spin_angle = spin_angle

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        angle[:] = np.deg2rad(self.spin_angle)

        return pos, angle


class SquareSpinIce(SpinIce):
    pass

class SquareSpinIceClosed(SquareSpinIce):
    def __init__(self, *, edge="symmetric", **kwargs):
        assert edge in ("symmetric", "asymmetric")
        self.edge = edge

        super().__init__(**kwargs)

    def _init_geometry(self):
        nx, ny = self.size
        sym = 1 if self.edge == "symmetric" else 0
        spin_count = (ny + sym) * nx + ny * (nx + sym)
        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        labels = []

        a = self.lattice_spacing
        y = 0
        i = 0
        for row in range(0, 2 * ny + sym):
            is_vert = row % 2 # 1 for vert, 0 for horiz
            ncols = nx
            x = 0

            if is_vert:
                # vertical row
                ncols += sym
            else:
                # horizontal row
                x += a/2

            for col in range(0, ncols):
                if is_vert:
                    angle[i] = np.pi/2
                pos[i] = [x, y]

                label = (row, col)
                labels.append(label)

                x += a
                i += 1

            y += a / 2

        self.labels = np.array(labels)

        return pos, angle

    def _init_spin(self, init):
        if isinstance(init, str) and init == 'ground':
            for row, col in self.labels:
                i = self.L[row, col]
                if row % 4 == 0 and col % 2 == 1:
                    self.spin[i] = -1
                elif row % 4 == 2 and col % 2 == 0:
                    self.spin[i] = -1
                elif row % 4 == 1 and col % 2 == 0:
                    self.spin[i] = -1
                elif row % 4 == 3 and col % 2 == 1:
                    self.spin[i] = -1
        else:
            super()._init_spin(init)

    _vertex_size = (3, 3)

    @property
    def _default_cell_size(self):
        return (self.lattice_spacing/2, self.lattice_spacing/2)

class SquareSpinIceOpen(SquareSpinIce):
    def __init__(self, *, neighbor_distance=sqrt(2), **kwargs):
        kwargs['neighbor_distance'] = neighbor_distance

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        angle[:] = np.pi/4

        for i in range(len(pos)):
            row, col = self.labels[i]
            if row % 2 == 0 and col % 2 == 1:
                angle[i] += np.pi/2
            elif row % 2 == 1 and col % 2 == 0:
                angle[i] += np.pi/2

        return pos, angle

    _vertex_size = (2, 2)

class PinwheelSpinIceDiamond(SquareSpinIceClosed):
    def __init__(self, *, spin_angle=45, neighbor_distance=10, **kwargs):
        self.spin_angle = spin_angle
        kwargs['neighbor_distance'] = neighbor_distance

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        # rotate each island by spin_angle
        angle[:] -= np.deg2rad(self.spin_angle)

        return pos, angle

class PinwheelSpinIceLuckyKnot(SquareSpinIceOpen):
    def __init__(self, *, spin_angle=45, neighbor_distance=10*sqrt(2), **kwargs):
        self.spin_angle = spin_angle
        kwargs['neighbor_distance'] = neighbor_distance

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        # rotate each island by spin_angle
        angle[:] -= np.deg2rad(self.spin_angle)

        return pos, angle

class PinwheelSpinIceRandom(PinwheelSpinIceDiamond):
    def __init__(self, *, spin_angle_disorder=0, **kwargs):
        self.spin_angle_disorder = spin_angle_disorder

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        # rotate each island by spin_angle
        std = self.spin_angle_disorder
        angle[:] += np.deg2rad(self.rng.normal(0, std, angle.shape))

        return pos, angle

class KagomeSpinIce(SpinIce):
    def _init_geometry(self):
        labels = []

        nx, ny = self.size

        n_top_bot = (2 * nx * 2)
        n_mid = (2 * nx + 1) * (ny - 1)
        n_vert = (nx + 1) * ny
        spin_count = n_top_bot + n_mid + n_vert

        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        n_rows = 2 * ny + 1
        last_row = n_rows - 1

        a = self.lattice_spacing
        y = 0
        i = 0
        for row in range(0, n_rows):
            if row % 2 == 0:
                # "horizontal" magnets (+/-30 degree)
                n_cols = 2 * nx + 1
                if row == 0 or row == last_row:
                    # first and last row has 1 less element
                    n_cols -= 1

                x = a/2
                col0 = 0
                if ny % 2 == 0 and row == last_row:
                    # even number of magnets, skip first magnet
                    x += a
                    col0 += 1

                for col in range(col0, col0 + n_cols):
                    pos[i] = [x, y]
                    angle[i] = np.deg2rad(30)
                    if row % 4 == 0 and col % 2 == 0:
                        angle[i] *= -1
                    if row % 4 == 2 and col % 2 == 1:
                        angle[i] *= -1

                    label = (row, col)
                    labels.append(label)

                    x += a
                    i += 1
            else:
                # vertical magnets (90 degrees)
                n_cols = nx + 1
                x = 0
                if row % 4 == 3:
                    x += a
                for col in range(0, n_cols):
                    pos[i] = [x, y]
                    angle[i] = np.deg2rad(90)

                    label = (row, col)
                    labels.append(label)

                    x += 2 * a
                    i += 1

            y += a * sqrt(3) / 2

        self.labels = np.array(labels)

        return pos, angle

    _vertex_size = (2,3)

    @property
    def _default_cell_size(self):
        sizex = self.lattice_spacing/2
        sizey = sqrt(3)*self.lattice_spacing/2
        return (sizex, sizey)

    def _init_spin(self, init):
        if isinstance(init, str) and init == 'ground':
            for row, col in self.labels:
                i = self.L[row, col]
                if row % 4 == 0 and(col-1)%6 < 3:
                    self.spin[i] = -1
                elif row % 4 == 2 and (col-1)%6 >=3:
                    self.spin[i] = -1
                elif row % 4 == 1 and col % 3 == 2:
                    self.spin[i] = -1
                elif row % 4 == 3 and col % 3 == 0:
                    self.spin[i] = -1

        else:
            super()._init_spin(init)

class KagomeSpinIceRotated(SpinIce):
    def _init_geometry(self):
        labels = []

        nx, ny = self.size

        n_top_bot = (2 * nx * 2)
        n_mid = (2 * nx + 1) * (ny - 1)
        n_vert = (nx + 1) * ny
        spin_count = n_top_bot + n_mid + n_vert

        pos = np.zeros((spin_count, 2), dtype=float)
        angle = np.zeros(spin_count, dtype=float)

        n_cols = 2 * nx + 1
        last_col = n_cols - 1

        a = self.lattice_spacing
        x = 0
        i = 0
        for col in range(0, n_cols):
            if col % 2 == 0:
                # "horizontal" magnets (+/-30 degree)
                n_rows = 2 * ny + 1
                if col == 0 or col == last_col:
                    # first and last col has 1 less element
                    n_rows -= 1

                y = a/2
                row0 = 0
                if nx % 2 == 0 and col == last_col:
                    # even number of magnets, skip first magnet
                    y += a
                    row0 += 1

                for row in range(row0, row0 + n_rows):
                    pos[i] = [x, y]
                    angle[i] = np.deg2rad(60)
                    if col % 4 == 0 and row % 2 == 0:
                        angle[i] *= -1
                    if col % 4 == 2 and row % 2 == 1:
                        angle[i] *= -1

                    label = (col, row)
                    labels.append(label)

                    y += a
                    i += 1
            else:
                # vertical magnets (90 degrees)
                n_rows = ny + 1
                y = 0
                if col % 4 == 3:
                    y += a
                for row in range(0, n_rows):
                    pos[i] = [x, y]
                    angle[i] = np.deg2rad(0)

                    label = (col, row)
                    labels.append(label)

                    y += 2 * a
                    i += 1

            x += a * sqrt(3) / 2

        self.labels = np.array(labels)

        return pos, angle

    _vertex_size = (3,2)

    @property
    def _default_cell_size(self):
        sizex = sqrt(3)*self.lattice_spacing/2
        sizey = self.lattice_spacing/2
        return (sizex, sizey)

    def _init_spin(self, init):
        if isinstance(init, str) and init == 'ground':
            for col, row in self.labels:
                i = self.L[col, row]
                if col % 2 == 0 and row % 3 == 2:
                    self.spin[i] = -1
                elif col % 4 == 1 and row % 3 != 2:
                    self.spin[i] = -1
                elif col % 4 == 3 and row % 3 != 0:
                    self.spin[i] = -1

        else:
            super()._init_spin(init)

class CustomSpinIce(SpinIce):
    def __init__(self, *, magnet_coords=[[1,1],[1,2]], magnet_angles=[0,90],
                 radians=False, coords_as_labels=False, labels=None, **kwargs):
        # load values from file
        if isinstance(magnet_coords, str):
            magnet_coords = np.array(read_table(magnet_coords))
        if isinstance(magnet_angles, str):
            magnet_angles = np.array(read_table(magnet_angles))
        if isinstance(labels, str):
            labels = np.array(read_table(labels))

        self.magnet_coords = magnet_coords
        if radians:
            self.magnet_angles = magnet_angles
        else:
            self.magnet_angles = np.deg2rad(magnet_angles)

        self.coords_as_labels = coords_as_labels
        self.custom_labels = labels

        super().__init__(**kwargs)

    def _init_geometry(self):
        pos = np.array(self.magnet_coords, dtype=float)
        angle = np.atleast_1d(np.array(self.magnet_angles, dtype=float).squeeze())

        assert pos.ndim == 2
        assert angle.ndim == 1
        assert pos.shape[0] == angle.shape[0], "unequal number of coordinates and angles"
        assert len(np.unique(pos,axis=0)) == len(pos), "magnet coordinates cannot be repeated"

        pos *= self.lattice_spacing

        if self.coords_as_labels:
            self.labels = self.magnet_coords.copy()
        else:
            self.labels = self.custom_labels

        return pos, angle

class LatticeSpinIce(SpinIce):
    def __init__(self, *, basis0=(1,0), basis1=(0,1), const_angle=0, radians=False, **kwargs):
        self.basis0 = basis0
        self.basis1 = basis1

        if not radians:
            const_angle = np.deg2rad(const_angle)
        self.const_angle = const_angle

        super().__init__(**kwargs)

    def _init_geometry(self):
        b0 = np.array(self.basis0)
        b1 = np.array(self.basis1)

        grid = np.array(np.meshgrid(np.arange(self.size[1]), np.arange(self.size[0]))).T.reshape(-1, 2)
        grid = np.fliplr(grid) #flip x and y so labels will be (row, col) or (b0, b1)
        pos = np.dot(grid, np.array([b0, b1]))
        angle = np.full(len(pos), self.const_angle)
        self.labels = grid

        pos = pos * self.lattice_spacing
        return pos, angle

    def set_basis(self, basis0=None, basis1=None):
        if basis0 is not None:
            self.basis0 = basis0
        if basis1 is not None:
            self.basis1 = basis1

        b0 = self.basis0
        b1 = self.basis1

        grid = self.labels
        pos = np.dot(grid, np.array([b0, b1]))
        pos = pos * self.lattice_spacing

        self.set_pos(pos)

class TileLatticeSpinIce(LatticeSpinIce):
    def __init__(self, *, angle_tile=((90,0),(0,90)), hole_tile=None, radians=False, **kwargs):
        if not radians:
            angle_tile = np.deg2rad(angle_tile)
        self.angle_tile = np.atleast_2d(angle_tile).T

        if hole_tile is not None:
            hole_tile = np.atleast_2d(hole_tile).T
            assert np.any(hole_tile), "Your hole_tile is all holes and no magnets!"
        self.hole_tile = hole_tile
        super().__init__(radians=radians, **kwargs)

    def _init_geometry(self):
        pos, angle = super()._init_geometry()

        if self.hole_tile is not None:
            grid = self.labels
            m, n = self.hole_tile.shape
            holes = self.hole_tile[np.mod(grid[:, 0], m), np.mod(grid[:, 1], n)]
            not_holes = np.nonzero(holes)
            pos = pos[not_holes]
            angle = angle[not_holes]
            self.labels = self.labels[not_holes]

        if self.angle_tile is not None:
            grid = self.labels
            m, n = self.angle_tile.shape
            angle = self.angle_tile[np.mod(grid[:, 0], m), np.mod(grid[:, 1], n)]

        return pos, angle

    def set_angle_tile(self, angle_tile, radians=True):
        grid = self.labels

        if angle_tile is not None:
            if not radians:
                angle_tile = np.deg2rad(angle_tile)
            self.angle_tile = np.atleast_2d(angle_tile).T
            m, n = self.angle_tile.shape
            angle = self.angle_tile[np.mod(grid[:, 0], m), np.mod(grid[:, 1], n)]
        else:
            self.angle_tile = angle_tile
            angle = np.full(self.spin_count, self.const_angle)

        self.set_angle(angle)



import numpy as np
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal

from flatspin import PinwheelSpinIceDiamond

def test_spin_angle():
    si = PinwheelSpinIceDiamond(size=(3,3))
    L = si.L
    l1 = L[::2]
    l2 = L[1::2]

    # Default spin_angle is 45 degrees clockwise
    assert_allclose(si.angle[l1], np.deg2rad(-45))
    assert_allclose(si.angle[l2], np.deg2rad(45))

    # 45 degrees counter-clockwise
    si = PinwheelSpinIceDiamond(size=(3,3), spin_angle=-45)
    assert_allclose(si.angle[l1], np.deg2rad(45))
    assert_allclose(si.angle[l2], np.deg2rad(90+45))

    # 30 degrees clockwise
    si = PinwheelSpinIceDiamond(size=(3,3), spin_angle=30)
    assert_allclose(si.angle[l1], np.deg2rad(-30))
    assert_allclose(si.angle[l2], np.deg2rad(90-30))

    # 33 degrees counter-clockwise
    si = PinwheelSpinIceDiamond(size=(3,3), spin_angle=-33)
    assert_allclose(si.angle[l1], np.deg2rad(33))
    assert_allclose(si.angle[l2], np.deg2rad(90+33))

@pytest.mark.parametrize("spin_angle", [45, -45, 30])
def test_find_vertices(spin_angle):
    si = PinwheelSpinIceDiamond(size=(3,3), spin_angle=spin_angle)
    L = si.L
    vi, vj, indices = si.find_vertices()

    assert_array_equal(vi, [0, 0, 0, 1, 1, 2, 2, 2, 3, 3, 4, 4, 4])
    assert_array_equal(vj, [0, 2, 4, 1, 3, 0, 2, 4, 1, 3, 0, 2, 4])

    assert tuple(indices[0]) == (L[0,0], L[1,0], L[1,1], L[2,0])
    assert tuple(indices[1]) == (L[0,1], L[1,1], L[1,2], L[2,1])
    assert tuple(indices[2]) == (L[0,2], L[1,2], L[1,3], L[2,2])

    assert tuple(indices[3]) == (L[1,1], L[2,0], L[2,1], L[3,1])
    assert tuple(indices[4]) == (L[1,2], L[2,1], L[2,2], L[3,2])

    assert tuple(indices[5]) == (L[2,0], L[3,0], L[3,1], L[4,0])
    assert tuple(indices[6]) == (L[2,1], L[3,1], L[3,2], L[4,1])
    assert tuple(indices[7]) == (L[2,2], L[3,2], L[3,3], L[4,2])

    assert tuple(indices[8]) == (L[3,1], L[4,0], L[4,1], L[5,1])
    assert tuple(indices[9]) == (L[3,2], L[4,1], L[4,2], L[5,2])

    assert tuple(indices[10]) == (L[4,0], L[5,0], L[5,1], L[6,0])
    assert tuple(indices[11]) == (L[4,1], L[5,1], L[5,2], L[6,1])
    assert tuple(indices[12]) == (L[4,2], L[5,2], L[5,3], L[6,2])

@pytest.mark.parametrize("spin_angle", [45, -45, 30])
def test_vertex_type(spin_angle):
    import matplotlib.pylab as plt

    si = PinwheelSpinIceDiamond(size=(2,2), edge='asymmetric', spin_angle=spin_angle)

    vertices = si.vertices()

    # first vertex sublattice
    v = vertices[0]

    # type 1
    si.spin[v] = [1, 1, -1, -1]
    assert si.vertex_type(v) == 1
    si.spin[v] = [-1, -1, 1, 1]
    assert si.vertex_type(v) == 1

    # type 2
    si.spin[v] = [1, 1, 1, 1]
    assert si.vertex_type(v) == 2
    si.spin[v] = [-1, 1, 1, -1]
    assert si.vertex_type(v) == 2
    si.spin[v] = [-1, -1, -1, -1]
    assert si.vertex_type(v) == 2
    si.spin[v] = [1, -1, -1, 1]
    assert si.vertex_type(v) == 2

    # type 3
    si.spin[v] = [-1, 1, 1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [-1, 1, -1, -1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [-1, -1, -1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, 1, -1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, -1, 1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, -1, -1, -1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [-1, -1, 1, -1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, 1, 1, -1]
    assert si.vertex_type(v) == 3

    # type 4
    si.spin[v] = [-1, 1, -1, 1]
    assert si.vertex_type(v) == 4
    si.spin[v] = [1, -1, 1, -1]
    assert si.vertex_type(v) == 4



    # second vertex sublattice
    v = vertices[1]

    # type 1
    si.spin[v] = [1, -1, 1, -1]
    assert si.vertex_type(v) == 1
    si.spin[v] = [-1, 1, -1, 1]
    assert si.vertex_type(v) == 1

    # type 2
    si.spin[v] = [1, 1, 1, 1]
    assert si.vertex_type(v) == 2
    si.spin[v] = [-1, 1, 1, -1]
    assert si.vertex_type(v) == 2
    si.spin[v] = [-1, -1, -1, -1]
    assert si.vertex_type(v) == 2
    si.spin[v] = [1, -1, -1, 1]
    assert si.vertex_type(v) == 2

    # type 3
    si.spin[v] = [-1, 1, 1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [-1, 1, -1, -1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [-1, -1, -1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, 1, -1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, -1, 1, 1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, -1, -1, -1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [-1, -1, 1, -1]
    assert si.vertex_type(v) == 3
    si.spin[v] = [1, 1, 1, -1]
    assert si.vertex_type(v) == 3

    # type 4
    si.spin[v] = [-1, -1, 1, 1]
    assert si.vertex_type(v) == 4
    si.spin[v] = [1, 1, -1, -1]
    assert si.vertex_type(v) == 4



from configparser import ConfigParser
cf = ConfigParser()
cf.read('../../setup.cfg')
console_scripts = cf['options.entry_points']['console_scripts']

for line in console_scripts.strip().splitlines():
    print(line)
    prog, mod = line.split('=')
    mod, func = mod.split(':')
    mod = mod.strip()
    prog = prog.strip()
    _, basename = mod.rsplit('.', 1)
    filename = f"cmdline/{basename}.rst"
    print("prog", prog, "mod", mod, "filename", filename)


    doc =  f".. _{prog}:\n\n"
    doc += f".. autoprogram:: {mod}:parser\n"
    doc += f"   :prog: {prog}\n"
    print(filename)
    print(doc)
    with open(filename, 'w') as fp:
        fp.write(doc)



# -*- coding: utf-8 -*-
"""
Check the health of a dataset.
"""
import os
import numpy as np
import pandas as pd
from flatspin.data import Dataset, read_table
from flatspin.cmdline import *

def main_fsck(dataset, args):
    if len(dataset) == 0:
        print(f"Dataset is empty, nothing to do")
        return 1

    missing = ~dataset.index['outdir'].apply(os.path.exists)
    missing = dataset.index[missing]

    if len(missing) > 0:
        print(f"Missing data for {len(missing)} runs")
        if args.verbose:
            pd.set_option('display.max_rows', None)
            pd.set_option('display.max_columns', None)
            pd.set_option('display.width', int(os.getenv('COLUMNS') if os.getenv('COLUMNS') is not None else None))
            print(missing)
        return 2
    else:
        print(f"Dataset appears healthy :)")
        return 0


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_argparser(__doc__)
parser.add_argument('-v', '--verbose', action='store_true')

def main():
    args = parser.parse_args()

    ds = main_dataset(args)

    return main_fsck(ds, args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import os
import argparse
import shlex
import numpy as np
import pandas as pd
import importlib
import inspect
import matplotlib.pyplot as plt

from flatspin import __version__
from flatspin.cmdline import StoreKeyValue, eval_params, main_version
from flatspin.data import Dataset, table_formats, is_archive_format, read_table, eval_dict
from flatspin.encoder import Encoder
from flatspin.runner import run, run_local, run_dist
from flatspin import runner
from flatspin.utils import get_default_params, import_class, get_subclasses

def main_worker(args):
    # Run!
    dataset = Dataset.read(args.basepath)

    worker_id = args.worker_id
    num_jobs = len(dataset)
    num_workers = args.num_workers

    # Calculate which jobs in the dataset to run
    # See https://stackoverflow.com/questions/27553218/algorithm-for-distributing-workload-in-a-thread-pool
    from_idx = (worker_id * num_jobs) // num_workers
    to_idx = ((worker_id + 1) * num_jobs) // num_workers

    dataset = dataset[from_idx:to_idx-1]
    run_local(dataset)

def main_list_params(args):
    model_class = import_class(args.model, 'flatspin.model')
    model_name = model_class.__name__
    encoder_class = import_class(args.encoder, 'flatspin.encoder')
    encoder_name = encoder_class.__name__

    print(f"Default model parameters [{model_name}]:")
    for k,v in get_default_params(model_class).items():
        print(f" {k}={v}")

    print("")

    print(f"Default run parameters:")
    for k,v in get_default_params(run).items():
        print(f" {k}={v}")

    print("")

    print(f"Default encoder parameters [{encoder_name}]:")
    for k,v in get_default_params(encoder_class).items():
        print(f" {k}={v}")

def main_list_encoders(args):
    module = args.encoder if args.encoder else 'flatspin.encoder'
    module = importlib.import_module(module)
    classes = get_subclasses(module, Encoder)

    print(f"Encoder listing for {module.__name__}:")
    for name, cls in classes:
        print(name)

def main_help_encoder(args):
    encoder_class = import_class(args.encoder, 'flatspin.encoder')
    print(f"{encoder_class.__module__}.{encoder_class.__name__}(", end='')
    print(", ".join(f"{k}={v}" for k,v in
        get_default_params(encoder_class).items()), end='')
    print(")", end='\n\n')
    if encoder_class.__doc__:
        print(inspect.cleandoc(encoder_class.__doc__))
    else:
        print("<no docstring>")

def visualize_run(model, i, max_i, result_table, params):
    model.plot(replace=True)
    plt.title(f"{i + 1} / {max_i}")
    plt.suptitle(params['outdir'])
    plt.pause(0.001)

def main_normal(args):
    model_class = import_class(args.model, 'flatspin.model')
    model_name = model_class.__name__
    encoder_class = import_class(args.encoder, 'flatspin.encoder')

    # -e Sine is really just an alias for -p encoder=Sine
    params = get_default_params(run)
    params['encoder'] = f'{encoder_class.__module__}.{encoder_class.__name__}'
    params.update(get_default_params(model_class))
    params.update(get_default_params(encoder_class))

    if args.params_file:
        prms = read_table(args.params_file)
        prms = dict(np.array(prms))
        prms = eval_dict(prms)
        params.update(prms)

    params.update(eval_params(args.param))

    info = {
        'model': f'{model_class.__module__}.{model_class.__name__}',
        'model_name': model_name,
        'data_format': args.format,
        'command': ' '.join(map(shlex.quote, sys.argv)),
        'version': __version__
    }

    basepath = args.basepath
    if os.path.exists(basepath):
        # Refuse to overwrite an existing dataset
        raise FileExistsError(basepath)
    os.makedirs(basepath)

    if is_archive_format(args.format):
        outdir = f"{model_name}.{args.format}"
    else:
        outdir = f"{model_name}.out"
    outpath = os.path.join(basepath, outdir)

    # Save dataset
    index = pd.DataFrame({'outdir': [outdir]})
    dataset = Dataset(index, params, info, basepath)
    dataset.save()

    # Run!
    if args.run == 'local':
        if args.show_plot:
            runner.run_sample_callback = visualize_run
        run_local(dataset)

    elif args.run == 'dist':
        run_dist(dataset)


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = argparse.ArgumentParser(description='Run flatspin simulation.')
parser.add_argument('-r', '--run', choices=['local', 'dist', 'none', 'worker'], default='local',
        help='run locally or distributed on a cluster')
parser.add_argument('-p', '--param', action=StoreKeyValue, default={},
        help='set model/run parameter key=value')
parser.add_argument('-m', '--model', default='SquareSpinIceClosed',
        help='name of model (default: %(default)s) ')
parser.add_argument('-e', '--encoder', default='Sine',
        help='input encoder (type of external field, default: %(default)s)')
parser.add_argument('-o', '--basepath',
        help='output directory for results')
parser.add_argument('-f', '--format', choices=table_formats.keys(), default='npz',
        help='format of output files (default: %(default)s)')
parser.add_argument('--params-file',
        help='load parameters from file (params.csv), loaded before -p')

parser.add_argument('--list-params', action='store_true')
parser.add_argument('--list-encoders', action='store_true')
parser.add_argument('--help-encoder', action='store_true')
parser.add_argument('--worker-id', help='worker id', type=int)
parser.add_argument('--num-workers', help='number of workers in total', type=int)
parser.add_argument('-V', '--version', help='display flatspin version', action='store_true')
parser.add_argument('--show-plot', help='plot the state as the simulation runs', action='store_true')

def main():
    args = parser.parse_args()

    if args.version:
        return main_version(args)
    if args.list_params:
        return main_list_params(args)

    if args.list_encoders:
        if args.encoder == parser.get_default('encoder'):
            args.encoder = 'flatspin.encoder'
        return main_list_encoders(args)

    if args.help_encoder:
        return main_help_encoder(args)

    if args.run == 'worker':
        if args.worker_id is None or args.num_workers is None or args.basepath is None:
            parser.error('The following arguments are required for worker mode: --worker-id, --num-workers, --basepath')
    else:
        if args.basepath is None:
            parser.error('The following arguments are required: -o/--basepath')

    if args.run == 'worker':
        # Taking part in distributed run
        return main_worker(args)
    elif args.list_params:
        return main_list_params(args)

    return main_normal(args)

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import pytest
import collections
from os import path

from flatspin.data import Dataset, read_csv
from flatspin.utils import eval_dict
from pandas.testing import assert_frame_equal

@pytest.fixture
def dataset(tmpdir_factory):
    basedir = tmpdir_factory.mktemp('test')
    index = pd.DataFrame({'A': np.arange(0,1,0.1), 'B': np.arange(0,10)//2, 'C': list('abcdefghij')})
    params = {'foo': 1, 'bar': [1,2,3]}
    info = {'array': np.array([3,2,1])}
    ds = Dataset(index, params, info, basedir)
    return ds

def test_len(dataset):
    assert len(dataset) == len(dataset.index)

def test_getitem(dataset):
    """ dataset[i] -> subset """
    ds = dataset[0]
    assert isinstance(ds, Dataset)
    assert len(ds) == 1

    ds = dataset[0:3]
    assert isinstance(ds, Dataset)
    assert len(ds) == 4

def test_iter(dataset):
    """ iter(dataset) -> iterator over subsets """
    ds = list(iter(dataset))
    assert len(ds) == len(dataset)
    assert all([isinstance(d, Dataset) for d in dataset])
    assert all([len(d) == 1 for d in dataset])

def test_keys(dataset):
    """ dataset.keys() -> list of indices """
    keys = dataset.keys()
    assert len(keys) == len(dataset)
    assert keys == list(range(len(dataset)))

def test_items(dataset):
    """ dataset.items() -> iterator over i, subsets """
    items = list(dataset.items())
    keys = [k for k,v in items]
    values = [v for k,v in items]

    assert keys == dataset.keys()
    assert values == list(dataset)

def test_drop_duplicates(dataset):
    ds = dataset.drop_duplicates(subset="A")
    assert isinstance(ds, Dataset)
    assert len(ds) == 10

    ds = dataset.drop_duplicates(subset=["B"])
    assert isinstance(ds, Dataset)
    assert len(ds) == 5

    ds = dataset.drop_duplicates(subset=["A","B"])
    assert isinstance(ds, Dataset)
    assert len(ds) == 10

    ds = dataset.drop_duplicates(subset="B").drop_duplicates(subset=["A"])
    assert isinstance(ds, Dataset)
    assert len(ds) == 5

    ds = dataset.drop_duplicates(subset=["B"])
    assert isinstance(ds, Dataset)
    assert len(ds) == 5

def test_filter(dataset):
    """ dataset.filter(A=0.5) -> subset """
    ds = dataset.filter(A=0.5)
    assert isinstance(ds, Dataset)
    assert len(ds) == 1

    ds = dataset.filter(A=1.0)
    assert isinstance(ds, Dataset)
    assert len(ds) == 0

    ds = dataset.filter(A=slice(None, 0.3))
    assert isinstance(ds, Dataset)
    assert len(ds) == 4

    ds = dataset.filter(A=slice(-100, 0.3))
    assert isinstance(ds, Dataset)
    assert len(ds) == 4

    ds = dataset.filter(A=slice(0.1, 0.3))
    assert isinstance(ds, Dataset)
    assert len(ds) == 3

    ds = dataset.filter(A=slice(0.4, None))
    assert isinstance(ds, Dataset)
    assert len(ds) == 6

    ds = dataset.filter(A=slice(0.4, 100))
    assert isinstance(ds, Dataset)
    assert len(ds) == 6

    ds = dataset.filter(A=slice(0.2, None, 2))
    assert isinstance(ds, Dataset)
    assert len(ds) == 4

    ds = dataset.filter(A=[0, 0.4, 0.9])
    assert isinstance(ds, Dataset)
    assert len(ds) == 3

    ds = dataset.filter(C='a')
    assert isinstance(ds, Dataset)
    assert len(ds) == 1

    ds = dataset.filter(C=('a', 'c', 'e'))
    assert isinstance(ds, Dataset)
    assert len(ds) == 3

    ds = dataset.filter(C=lambda x: True)
    assert isinstance(ds, Dataset)
    assert len(ds) == 10

    ds = dataset.filter(C=lambda x: False)
    assert isinstance(ds, Dataset)
    assert len(ds) == 0

    ds = dataset.filter(C=lambda x: ord(x)>103)
    assert isinstance(ds, Dataset)
    assert len(ds) == 3

def test_filter_big(dataset):
    dataset.index["big_int"] = np.arange(10000, 10000 + 10)
    dataset.index["big_float"] = np.arange(10000.0, 10000.0 + 10)
    # don't round on int
    ds = dataset.filter(big_int=10001)
    assert len(ds) == 1
    ds = dataset.filter(big_float=10001)
    assert len(ds) == 3

def test_filter_small(dataset):
    print(len(dataset.index))
    dataset.index['small'] = np.repeat([1e-24, 1.1e-24, 1.01e-24, 1.001e-24, 0], 2)
    # size of target value determines precision of match
    print(dataset)
    ds = dataset.filter(small=1e-24)
    assert len(ds) == 2
    ds = dataset.filter(small=1.1e-24)
    assert len(ds) == 2
    ds = dataset.filter(small=1.01e-24)
    assert len(ds) == 2
    ds = dataset.filter(small=1.001e-24)
    assert len(ds) == 2
    ds = dataset.filter(small=0)
    assert len(ds) == 2


def test_groupby(dataset):
    """ dataset.groupby(column) -> iterator """
    groups = list(dataset.groupby('B'))
    B = np.unique(dataset.index['B'])
    for i, ds in groups:
        assert np.all(ds.index['B'] == B[i])

def test_save(dataset):
    dataset.save()
    assert path.exists(path.join(dataset.basepath, 'index.csv'))
    assert path.exists(path.join(dataset.basepath, 'params.csv'))
    assert path.exists(path.join(dataset.basepath, 'info.csv'))

    index = read_csv(path.join(dataset.basepath, 'index.csv'))
    assert_frame_equal(index, dataset.index, check_dtype=False)

    params = read_csv(path.join(dataset.basepath, 'params.csv'))
    params = dict(np.array(params))
    params = eval_dict(params)
    assert params == dataset.params

    info = read_csv(path.join(dataset.basepath, 'info.csv'))
    info = dict(np.array(info))
    info = eval_dict(info)
    assert np.all(info['array'] == dataset.info['array'])

def test_read(dataset):
    dataset.save()
    dataset2 = Dataset.read(dataset.basepath)

    assert_frame_equal(dataset.index, dataset2.index, check_dtype=False)
    assert dataset.params == dataset2.params
    assert np.all(dataset.info['array'] == dataset2.info['array'])



import numpy as np
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal
from itertools import groupby

from flatspin import TileLatticeSpinIce

def test_init(opencl):
    b0, b1 = [1,0], [0,1]
    a=0.5
    hole_tile = [(1,0), (1,1)]
    angle_tile = (0, 90)
    size = (3,3)
    si = TileLatticeSpinIce(size=size, lattice_spacing=a, basis0=b0, basis1=b1, const_angle=45, radians=False,
                             hole_tile=hole_tile, angle_tile=angle_tile, **opencl)

    assert si.size == (3,3)
    assert si.spin_count == 7
    assert si.num_neighbors == 3

    assert_array_equal(si.pos, np.array([[0. , 0. ],
                                         [1. , 0. ],
                                         [0. , 0.5],
                                         [0.5, 0.5],
                                         [1. , 0.5],
                                         [0. , 1. ],
                                         [1. , 1. ]]))
    assert_array_equal(si.angle, [0, 0, 0, np.deg2rad(90), 0, 0, 0])

def test_geometry():
    #test square lattice
    b0, b1 = [1,0], [0,1]
    angle_tile = ((90 + 45, 45), (45, 90 + 45))
    si = TileLatticeSpinIce(size=(3,3), basis0=b0, basis1=b1, radians=False, angle_tile=angle_tile)
    assert_array_equal(si.pos, np.array([[0, 0],
                                         [1, 0],
                                         [2, 0],
                                         [0, 1],
                                         [1, 1],
                                         [2, 1],
                                         [0, 2],
                                         [1, 2],
                                         [2, 2]]))
    assert_array_equal(si.angle, np.deg2rad([135, 45, 135, 45, 135, 45, 135, 45, 135]))

    #test kagome lattice
    b0, b1 = [1,0], [0.5, np.sin(np.deg2rad(60))]
    angle_tile = (30, -30), (90, 90)
    hole_tile = ((1,1),(1,0))
    si = TileLatticeSpinIce(size=(4,3), basis0=b0, basis1=b1, radians=False, angle_tile=angle_tile, hole_tile=hole_tile)
    sin60 = np.sin(np.deg2rad(60))
    assert_array_equal(si.pos, np.array([[0, 0],
                                         [1, 0],
                                         [2, 0],
                                         [3, 0],
                                         [0.5, sin60],
                                         [2.5, sin60],
                                         [1, 2 * sin60],
                                         [2, 2 * sin60],
                                         [3, 2 * sin60],
                                         [4, 2 * sin60]]))
    assert_array_equal(si.angle, np.deg2rad([30, -30, 30, -30, 90, 90, 30, -30, 30, -30]))


def test_labels():
    si = TileLatticeSpinIce(size=(2,2))
    assert_array_equal(si.labels,
            [[0, 0],[1, 0],
             [0, 1],[1, 1]])
    si = TileLatticeSpinIce(size=(3,3), hole_tile=(1, 0))
    assert_array_equal(si.labels,
            [[0, 0],
             [2, 0],
             [0, 1],
             [2, 1],
             [0, 2],
             [2, 2]])


def test_indexof():
    si = TileLatticeSpinIce(size=(4,4), hole_tile=((1,1),(0,0)))

    inds = [si.indexof(tuple(l)) for l in si.labels]
    assert_array_equal(inds, si.indices())

def test_set_angle_tile(opencl):
    b0, b1 = [1,0], [0,1]
    a=0.5
    hole_tile = [(1,0), (1,1)]
    angle_tile = (0, 90)
    size = (3,3)
    si = TileLatticeSpinIce(size=size, lattice_spacing=a, basis0=b0, basis1=b1, const_angle=45, radians=False,
                             hole_tile=hole_tile, angle_tile=angle_tile, **opencl)

    assert si.size == (3,3)
    assert si.spin_count == 7
    assert si.num_neighbors == 3

    assert_array_equal(si.angle, [0, 0, 0, np.deg2rad(90), 0, 0, 0])

    angle_tile = (15, 30)
    si.set_angle_tile(angle_tile, radians=False)

    a15 = np.deg2rad(15)
    a30 = np.deg2rad(30)
    assert_array_equal(si.angle, [a15, a15, a15, a30, a15, a15, a15])

    angle_tile = (a30, a15)
    si.set_angle_tile(angle_tile)
    assert_array_equal(si.angle, [a30, a30, a30, a15, a30, a30, a30])


    si.set_angle_tile(None)
    assert_array_equal(si.angle, si.const_angle)

def test_num_neighbors():
    size = (10,10)

    # nearest neighborhood (default)
    si = TileLatticeSpinIce(size=size, hole_tile=(1,0))
    assert si.num_neighbors == 2

    # global neighborhood
    si = TileLatticeSpinIce(size=size, hole_tile=(1,0), neighbor_distance=np.inf)
    assert si.num_neighbors == si.spin_count - 1


def test_neighbors():
    si = TileLatticeSpinIce(size=(6,6), basis0=[0.1,0.9], basis1=[0.9, 0], angle_tile=((0,45,90),(45,90,0)),
                            hole_tile=((1,1),(1,0),(0,1)))
    L = si.L

    # First "horizontal" row
    i = L[0,0]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 0], L[0, 1]}

    i = L[0,1]
    ns = set(si.neighbors(i))
    assert ns == {L[0, 0]}

    # First vertical row
    i = L[1,0]
    ns = set(si.neighbors(i))
    assert ns == {L[0, 0], L[2, 0]}

    i = L[1,2]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 3]}

    # Middle "horizontal" row
    i = L[2,0]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 0], L[2, 1], L[3, 0]}

    i = L[2,3]
    ns = set(si.neighbors(i))
    assert ns == {L[1, 3], L[3, 3], L[2, 4]}


def test_spin_dipolar_field(opencl):
    # default lattice spacing = 1 such that NN islands have coupling strength 1.5
    si = TileLatticeSpinIce(size=(6,6), basis0=[0.1,0.9], basis1=[0.9, 0], angle_tile=((0,45,90),(45,90,0)),
                            hole_tile=((1,1),(1,0),(0,1)))
    L = si.L

    a, b = 0.1767767, 0.53033009
    assert si.spin_dipolar_field(L[0,0], L[0,1]) == approx([1.9399363, -0.96996815])
    assert si.spin_dipolar_field(L[0,0], L[1,0]) == approx([-0.6038843, 2.1832739])

    assert si.spin_dipolar_field(L[0,0], L[2,0]) == approx([0.0554293, 0.3305227])

    assert si.spin_dipolar_field(L[1,0], L[0,0]) == approx([-0.6038843, 1.2309949])
    assert si.spin_dipolar_field(L[1,0], L[0,1]) == approx([-0.5668027, 0.100721])

    assert si.spin_dipolar_field(L[1,2], L[0,1]) == approx([0.8179162, -0.0646619])
    assert si.spin_dipolar_field(L[1,2], L[2,1]) == approx([-0.4720105, -0.734549])
    assert si.spin_dipolar_field(L[1,2], L[0,0]) == approx([0.1986713, -0.0220363])
    assert si.spin_dipolar_field(L[1,2], L[2,0]) == approx([-0.1573542, 0.0891495])

def test_dipolar_field(opencl):
    si = TileLatticeSpinIce(size=(6,6), basis0=[0.1,0.9], basis1=[0.9, 0], angle_tile=((0,45,90),(45,90,0)),
                            hole_tile=((1,1),(1,0),(0,1)), lattice_spacing=0.3, **opencl)
    L = si.L

    assert si.dipolar_field(L[2,0]) == approx([0.09728548, -0.03119054])
    si.flip(L[3,0])
    assert si.dipolar_field(L[2,0]) == approx([0.06443851, -0.12729835])
    si.flip(L[2,1])
    assert si.dipolar_field(L[2,0]) == approx([0.06443851, 0.0759227])

def test_external_field(opencl):
    si = TileLatticeSpinIce(size=(6,6), basis0=[0.1,0.9], basis1=[0.9, 0], angle_tile=((0,45,90),(45,90,0)),
                            hole_tile=((1,1),(1,0),(0,1)), lattice_spacing=0.3, **opencl)
    L = si.L
    a30 = np.deg2rad(30)
    cos30 = np.cos(a30)
    sin30 = np.sin(a30)
    cos60 = np.cos(np.deg2rad(60))
    sin60 = np.sin(np.deg2rad(60))

    assert si.external_field(L[0,0]) == approx([0, 0])
    assert si.external_field(L[1,0]) == approx([0, 0])

    # Horizontal
    si.set_h_ext((1,0))
    assert si.external_field(L[0,0]) == approx([1.0, 0.0])
    assert si.external_field(L[0,1]) == approx([0.7071068, -0.7071068])
    assert si.external_field(L[1,0]) == approx([0.7071068, -0.7071068])

    # Vertical
    si.set_h_ext((0,1))
    assert si.external_field(L[0,0]) == approx([0.0, 1.0])
    assert si.external_field(L[0,1]) == approx([0.7071068, 0.7071068])
    assert si.external_field(L[1,0]) == approx([0.7071068, 0.7071068])

    # 30 degrees
    si.set_h_ext((cos30, sin30))
    assert si.external_field(L[0,0]) == approx([0.8660254, 0.5])
    assert si.external_field(L[0,1]) == approx([0.9659258, -0.258819])
    assert si.external_field(L[1,0]) == approx([0.9659258, -0.258819])

    # -30 degrees
    si.set_h_ext((np.cos(-a30), np.sin(-a30)))
    assert si.external_field(L[0,0]) == approx([0.8660254, -0.5])
    assert si.external_field(L[0,1]) == approx([0.258819, -0.9659258])
    assert si.external_field(L[1,0]) == approx([0.258819, -0.9659258])

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_flippable(opencl, switching):
    si = TileLatticeSpinIce(size=(6,6), basis0=[0.1,0.9], basis1=[0.9, 0], angle_tile=((0,45,90),(45,90,0)),
                            hole_tile=((1,1),(1,0),(0,1)), switching=switching, lattice_spacing=0.3, alpha=0.01, **opencl)
    if switching == "budrikis":
        assert len(si.flippable()) == 3
    elif switching == "sw":
        assert len(si.flippable()) == 5

    si.flip(si.L[1,2])
    E = si.switching_energy()
    flippable = [i for i in si.indices() if E[i] > 0]
    assert len(flippable) > 0
    assert_array_equal(si.flippable(), flippable)

    si.polarize()
    si.set_h_ext([-20, -20])
    assert len(si.flippable()) == si.spin_count


def test_spin_grid(opencl):
    si = TileLatticeSpinIce(size=(3,3), basis0=[-0.1,0.9], basis1=[0.9, -0.1], angle_tile=((0,45),(90,0)),
                            hole_tile=((0,1)), **opencl)
    # grid_spacing == lattice_spacing
    grid = si.spin_grid()
    assert grid.shape == (1, 3, 2)
    expect = [[[0.70710678, 0.70710678], 
              [1, 0], 
              [0.70710678, 0.70710678]]]

    assert_allclose(grid, np.array(expect))



#!/usr/bin/env python

import setuptools
# Workaround for https://github.com/pypa/pip/issues/7953
import site
import sys
site.ENABLE_USER_SITE = "--user" in sys.argv[1:]

if __name__ == "__main__":
    setuptools.setup()



from os import path
import numpy as np
from numpy.linalg import norm
import matplotlib.pyplot as plt
from matplotlib import ticker
from matplotlib.colors import Normalize, ListedColormap, LinearSegmentedColormap
from matplotlib.quiver import Quiver
import matplotlib.cm
import matplotlib.animation as animation
from matplotlib.path import Path
from matplotlib import _api
from scipy.spatial import cKDTree
import warnings

fs_colours= [      [.9, .1, .1],
                    [.9, .9, .1],
                    [.1, .7, .2],
                    [.1, .7, .9],
                    [.1, .1, .9],
                    [.9, .1 , .9],
                    [.9, .1, .1]]
matplotlib.cm.unregister_cmap('flatspin')
matplotlib.cm.register_cmap('flatspin', LinearSegmentedColormap.from_list('flatspin', fs_colours, N=1000))
normalize_rad = Normalize(vmin=0, vmax=2*np.pi)

# Colormap from Li et al., (2019)
clist = list('ckkbbkkggkkrrkkc')
clist = ['#00da00' if c == 'g' else c for c in clist]
clist = ['#0800da' if c == 'b' else c for c in clist]
clist = ['#ed0912' if c == 'r' else c for c in clist]
clist = ['#00ccff' if c == 'c' else c for c in clist]

cmap = ListedColormap(clist)
matplotlib.cm.unregister_cmap('li2019')
matplotlib.cm.register_cmap('li2019', cmap)

# Type 1: green, type 2: blue, type 3: red, type 4: gray
clist = [(.1, .7, .2), (.1, .1, .9), (.9, .1, .1), (.4,.4,.4)]
cmap = ListedColormap(clist)
matplotlib.cm.unregister_cmap('vertex-type')
matplotlib.cm.register_cmap('vertex-type', cmap)

def quadrilaterals(X):
    dX = X[1:]-X[:-1]
    dX = np.pad(dX, 1, mode='edge')
    dX[:-1] *= -1 # subtract dX/2 from all but last element, where we add dX/2
    # quads needs one more value (see pcolormesh)
    quads = np.pad(X, (0,1), mode='edge').astype(float)
    quads += dX/2
    return quads

def format_label(label, format_spec='{:g}'):
    if isinstance(label, tuple):
        lbl = ', '.join(map(format_label, label))
        return f'({lbl})'

    if isinstance(label, list):
        return ', '.join(map(format_label, label))

    try:
        return format_spec.format(label)
    except:
        return str(label)

def format_labels(labels, format_spec='{:g}'):
    return [format_label(l, format_spec) for l in labels]

def heatmap2(X, Y, Z, xlabel='x', ylabel='y', zlabel='z', xlabels=None, ylabels=None, **kwargs):
    X = np.array(X)
    Y = np.array(Y)
    Z = np.array(Z)
    print(X.shape, Y.shape, Z.shape)
    print(X.dtype, Y.dtype, Z.dtype)

    if not xlabels:
        xlabels = format_labels(X)

    #if X.dtype == object:
    #    X = np.arange(len(X))

    if not ylabels:
        ylabels = format_labels(Y)

    #if Y.dtype == object:
    #    Y = np.arange(len(Y))

    Xi = np.arange(len(X))
    Yi = np.arange(len(Y))
    print(Yi)
    qX = quadrilaterals(Xi)
    qY = quadrilaterals(Yi)

    def format_coord(x, y):
        # find nearest x,y
        #idx = np.abs(x-Xi).argmin()
        #idy = np.abs(y-Yi).argmin()
        idx = int(x + 0.5)
        idy = int(y + 0.5)
        x = xlabels[idx]
        y = ylabels[idy]
        z = Z[idy, idx]
        return f'{xlabel}={x}, {ylabel}={y}, {zlabel}={z:g}'

    ax = plt.gca()
    ax.format_coord = format_coord
    ax.xaxis.set_major_locator(ticker.FixedLocator(Xi, nbins=min(len(Xi), 10)))
    ax.yaxis.set_major_locator(ticker.FixedLocator(Yi, nbins=min(len(Yi), 10)))
    ax.xaxis.set_major_formatter(ticker.FuncFormatter((lambda x,pos: xlabels[x])))
    ax.yaxis.set_major_formatter(ticker.FuncFormatter((lambda y,pos: ylabels[y])))
    #ax.xaxis.set_major_formatter(ticker.FuncFormatter((lambda x,pos: (x,pos))))
    #ax.yaxis.set_major_formatter(ticker.FuncFormatter((lambda y,pos: (y,pos))))
    #ax.set_xticklabels(xlabels)
    #ax.set_yticklabels(ylabels)
    return plt.pcolormesh(qX, qY, Z, **kwargs)

def heatmap(X, Y, Z, xlabel='x', ylabel='y', zlabel='z',
        xlim=None, ylim=None, zlim=None, nticks=10,
        xlabel_format='{:g}', ylabel_format='{:g}',
        ax=None,
        **kwargs):

    xlabels = format_labels(X, xlabel_format)
    ylabels = format_labels(Y, ylabel_format)

    X = np.array(X)
    Y = np.array(Y)
    Z = np.array(Z)

    Xi = np.arange(len(X))
    Yi = np.arange(len(Y))

    def format_coord(x, y):
        try:
            idx = int(x)
            idy = int(y)
            x = xlabels[idx]
            y = ylabels[idy]
            z = Z[idy, idx]
            return f'{xlabel}={x}, {ylabel}={y}, {zlabel}={z:g}'
        except IndexError:
            return ''

    if ax is None:
        ax = plt.gca()

    ax.format_coord = format_coord
    nbins = min(len(Xi), nticks) if X.dtype != 'object' else None
    ax.xaxis.set_major_locator(ticker.FixedLocator(Xi+.5, nbins=nbins))
    nbins = min(len(Yi), nticks) if Y.dtype != 'object' else None
    ax.yaxis.set_major_locator(ticker.FixedLocator(Yi+.5, nbins=nbins))
    ax.xaxis.set_major_formatter(ticker.FuncFormatter((lambda x,pos: xlabels[int(x)])))
    ax.yaxis.set_major_formatter(ticker.FuncFormatter((lambda y,pos: ylabels[int(y)])))

    if zlim:
        kwargs['vmin'], kwargs['vmax'] = zlim

    pcolor = ax.pcolormesh(Z, **kwargs)

    # We have integer indexes on the axes, so need to map x values to x indexes
    if xlim and X.dtype != 'object':
        xmin, xmax = xlim
        ximin = np.argmax(X >= xmin)
        ximax = np.argmax(X > xmax)
        ax.set_xlim(ximin, ximax)

    if ylim and Y.dtype != 'object':
        ymin, ymax = ylim
        yimin = np.argmax(Y >= ymin)
        yimax = np.argmax(Y > ymax)
        ax.set_ylim(yimin, yimax)

    return pcolor

def vector_colors(U, V):
    C = np.arctan2(V, U) # color
    C[C<0] = 2*np.pi + C[C<0]
    return C

def mask_zero_vectors(UV):
    # mask out zero vectors
    mag = norm(UV, axis=-1)
    mask = np.repeat(mag <= 1e-12, 2).reshape(UV.shape)
    return np.ma.array(UV, mask=mask)

class SpinQuiver(Quiver):
    """ Custom Quiver with support for different spin styles (shapes) """

    def __init__(self, *args, style="arrow", **kwargs):
        assert style in ("arrow", "rectangle", "stadium")
        self.style = style
        super().__init__(*args, **kwargs)

    def _h_arrows(self, length):
        # let Quiver handle the "arrow" style
        if self.style == "arrow":
            return super()._h_arrows(length)

        # modified from Quiver._h_arrows()
        N = len(length)
        length = length.reshape(N, 1)

        # This number is chosen based on when pixel values overflow in Agg
        # causing rendering errors
        # length = np.minimum(length, 2 ** 16)
        np.clip(length, 0, 2 ** 16, out=length)

        # Draw a rectangle
        X = np.array([0, 0, 1, 1, 0], np.float64) * length
        Y = np.array([0, 1, 1, 0, 0], np.float64) - 0.5
        Y = np.repeat(Y[np.newaxis, :], N, axis=0)

        if self.pivot == 'middle':
            X -= 0.5 * X[:, 3, np.newaxis]
        elif self.pivot == 'tip':
            # numpy bug? using -= does not work here unless we multiply by a
            # float first, as with 'mid'.
            X = X - X[:, 3, np.newaxis]
        elif self.pivot != 'tail':
            _api.check_in_list(["middle", "tip", "tail"], pivot=self.pivot)

        tooshort = length < self.minlength
        if tooshort.any():
            # Use a square
            tooshort = tooshort.ravel()
            X[tooshort] = np.array([-1, 0, 1, 0, -1]) * self.minlength * 0.5
            Y[tooshort] = np.array([ 0, 1, 0, -1, 0]) * self.minlength * 0.5

        # Mask handling is deferred to the caller, _make_verts.
        return X, Y

    def set_verts(self, verts, closed=True):
        # The drawn Paths are created in set_verts().
        # We override this to create paths for stadium shape.
        if self.style != "stadium":
            return super().set_verts(verts, closed)

        if len(verts) == 0:
            return

        verts = np.array(verts)
        assert verts.shape[1] == 5

        codes = [Path.MOVETO,
                 Path.CURVE4,
                 Path.CURVE4,
                 Path.CURVE4,
                 Path.LINETO,
                 Path.CURVE4,
                 Path.CURVE4,
                 Path.CURVE4,
                 Path.LINETO]

        # New vertices for the path
        new_verts = np.empty((len(verts), len(codes), 2))

        # Create left bezier curve
        v1 = .7 * (verts[:,1] - verts[:,0])
        # n1 is v1 rotated by 90 degrees
        n1 = np.column_stack([-v1[:,1], v1[:,0]])

        # First and second control points
        new_verts[:,1] = verts[:,0] + .25*n1
        new_verts[:,2] = verts[:,1] + .25*n1

        # Move original vertices back to accomodate curve
        new_verts[:,0] = verts[:,0] - .75*n1
        new_verts[:,3] = verts[:,1] - .75*n1

        # Create right bezier curve in similar fashin
        v2 = .7 * (verts[:,3] - verts[:,2])
        n2 = np.column_stack([-v2[:,1], v2[:,0]])
        new_verts[:,5] = verts[:,2] + .25*n2
        new_verts[:,6] = verts[:,3] + .25*n2
        new_verts[:,4] = verts[:,2] - .75*n2
        new_verts[:,7] = verts[:,3] - .75*n2

        # Close the path
        new_verts[:,8] = new_verts[:,0]

        # Create path objects
        self._paths = [Path(xy, codes) for xy in new_verts]


def spin_quiver(*args, ax=None, **kwargs):
    # Identical to ax.quiver(), but for SpinQuiver
    ax = ax or plt.gca()
    args = ax._quiver_units(args, kwargs)
    q = SpinQuiver(ax, *args, **kwargs)
    ax.add_collection(q, autolim=True)
    ax._request_autoscale_view()
    return q

def plot_vectors(XY, UV, C=None, style="arrow", mask_zero=True, replace=False,
        normalize=False, ax=None, **kwargs):

    if 'arrows' in kwargs:
        # Maintain backwards compatibility with old arrows=True/False
        warnings.warn('arrows is deprecated, use style="arrow" instead.',
                DeprecationWarning)
        arrows = kwargs.pop('arrows')
        style = "arrow" if arrows else "rectangle"

    XY = np.atleast_2d(XY)
    UV = np.atleast_2d(UV)
    XY = XY.reshape((-1, 2)) # quiver wants 2D XY
    X = XY[...,0] # x components
    Y = XY[...,1] # y components

    if normalize:
        # Normalize vectors to unit length
        nmax = norm(UV.reshape((-1,2)), axis=-1).max()
        if nmax != 0:
            UV = UV / nmax

    if mask_zero:
        UV = mask_zero_vectors(UV)

    U = UV[..., 0] # x components
    V = UV[..., 1] # y components
    if C is None:
        C = vector_colors(U, V)
        kwargs.setdefault('norm', normalize_rad)
    #C = np.arctan2(V, U) # color
    #C[C<0] = 2*np.pi + C[C<0]

    if ax is None:
        ax = plt.gca()

    if replace and len(ax.collections) > 0:
        quiv = ax.collections[0]
        old_XY = quiv.get_offsets()
        if not np.array_equal(old_XY, XY):
            # Positions have changed, create a new quiver
            replace = False
            quiv.remove()

    if replace and len(ax.collections) > 0:
        # Replace current quiver data
        quiv = ax.collections[0]
        quiv.set_UVC(U, V, C)

    else:
        # Make a new quiver

        # Determine scale based on the average distance between nearest neighbors
        tree = cKDTree(XY)
        min_dist = np.mean(tree.query(XY, k=[2])[0])
        scale = 1.15/min_dist if not np.isinf(min_dist) and min_dist != 0 else 1

        kwargs.setdefault('cmap', 'flatspin')
        kwargs.setdefault('scale', scale)
        kwargs.setdefault('pivot', 'middle')
        if style == "arrow":
            kwargs.setdefault('width', 0.2/scale)
            kwargs.setdefault('headwidth', 3)
            kwargs.setdefault('headlength', 2)
            kwargs.setdefault('headaxislength', 2)
        else:
            # rectangle / stadium
            kwargs.setdefault('width', 0.3/scale)
            kwargs.setdefault('headwidth', 0)
            kwargs.setdefault('headlength', 0)
            kwargs.setdefault('headaxislength', 0)

        if type(kwargs["cmap"]) is str and kwargs["cmap"].startswith("peem"):
            # clever peem colormap
            peem_angle = np.deg2rad(float(kwargs["cmap"].strip("peem")))
            intensity = np.cos(np.linspace(0, 2 * np.pi, 360) - peem_angle)
            intensity = intensity * 0.5 + 0.5  # scale from (-1,1) to (0,1)
            intensity = np.tile(intensity, (3, 1)).T  # [[0],...,[1]] -> [[0,0,0],...,[1,1,1]]
            kwargs["cmap"] = ListedColormap(intensity)
            ax.set_facecolor([0.5] * 3)

        # If user passed in explicit colors (via the 'color' kwarg) we omit C
        # from args, since quiver doesn't allow C to be passed as kwarg.
        args = [X, Y, U, V]
        if 'color' not in kwargs:
            args.append(C)
        quiv = spin_quiver(*args, ax=ax, style=style, units='xy',
                angles='xy', scale_units='xy', **kwargs)

        ax.set_aspect('equal')

        # Make space for the whole arrow
        xmin, xmax = np.min(X), np.max(X)
        ymin, ymax = np.min(Y), np.max(Y)
        ax.set_xlim(xmin - 1/scale, xmax + 1/scale)
        ax.set_ylim(ymin - 1/scale, ymax + 1/scale)

    return quiv

def plot_vector_image(XY, UV, mask_zero=True, replace=False, cmap='flatspin', ax=None):
    if mask_zero:
        UV = mask_zero_vectors(UV)

    U = UV[...,0] # x components
    V = UV[...,1] # y components
    C = vector_colors(U, V)

    cmap = plt.get_cmap(cmap)
    mag = norm(UV, axis=-1)
    alpha = mag

    # print(f'alpha {np.min(alpha)} {np.max(alpha)}')
    im = cmap(normalize_rad(C))
    im[:,:,-1] = alpha
    # print(f'im {im.shape} {im}')

    if ax is None:
        ax = plt.gca()

    if replace and len(ax.images) > 0:
        # Replace current image data
        ax.images[0].set_data(im)
        return ax.images[0]

    X = np.unique(XY[..., 0]) # x components
    Y = np.unique(XY[..., 1]) # y components
    Xi = np.arange(len(X))
    Yi = np.arange(len(Y))

    xmin, xmax = X.min(), X.max()
    ymin, ymax = Y.min(), Y.max()

    def format_coord(x, y):
        x = xmin + (xmax-xmin) * x / (im.shape[1] - 1)
        y = ymin + (ymax-ymin) * y / (im.shape[0] - 1)
        return f'x={x:g}, y={y:g}'

    ax.format_coord = format_coord
    nbins = min(len(Xi), 10)
    ax.xaxis.set_major_locator(ticker.FixedLocator(Xi, nbins=nbins))
    nbins = min(len(Yi), 10)
    ax.yaxis.set_major_locator(ticker.FixedLocator(Yi, nbins=nbins))
    ax.xaxis.set_major_formatter(ticker.FuncFormatter((lambda x,pos: "{:g}".format(X[int(x)]))))
    ax.yaxis.set_major_formatter(ticker.FuncFormatter((lambda y,pos: "{:g}".format(Y[int(y)]))))

    return ax.imshow(im, origin='lower')

def montage_fig(n_axes, title=False):
    """ Create a figure with n_axes subplots on a grid """
    n_rows = int(np.ceil(np.sqrt(n_axes)))
    n_cols = int(np.ceil(n_axes / n_rows))
    gs = dict(wspace=0.05, hspace=0.25, left=0.05, right=0.95, top=0.95, bottom=0.05)
    figsize = [6.4, 6.4]

    if title:
        gs.update(top=0.85) # todo: fixme

    fig, axes = plt.subplots(n_rows, n_cols, gridspec_kw=gs, figsize=figsize, squeeze=False)
    axes = axes.flatten()

    for ax in axes:
        ax.tick_params(
            top=False, bottom=False, left=False, right=False,
            labeltop=False, labelbottom=False, labelleft=False, labelright=False)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_visible(False)
        ax.spines['left'].set_visible(False)

        ax.grid(False)

    # hide unused axes
    for ax in axes[n_axes:]:
        ax.set_visible(False)

    return fig, axes

def vector_montage(axes, positions, vectors, labels=None, style="image", **kwargs):
    for i, (XY, UV, ax) in enumerate(zip(positions, vectors, axes)):
        if style == "image":
            plot_vector_image(XY, UV, replace=True, ax=ax, **kwargs)
        else:
            plot_vectors(XY, UV, style=style, replace=True, ax=ax, **kwargs)

        if labels:
            ax.set_title(labels[i], size='x-small', pad=2)

def save_animation(ani, outfile, fps=30, dpi=100):
    base, ext = path.splitext(outfile)
    ext = ext.lstrip('.')

    if ext == 'mp4':
        print(f"Saving video {outfile}...")
        Writer = animation.writers['ffmpeg']
        writer = Writer(fps=fps, codec='h264', extra_args=['-crf', '0', '-g', '1'])
        ani.save(outfile, writer=writer, dpi=dpi)
        return

    if ext in animation.writers['imagemagick_file'].supported_formats:
        print(f"Saving image(s) {outfile}...")
        Writer = animation.writers['imagemagick_file']
        writer = Writer(fps=fps, extra_args=['-trim', '+repage'])
        ani.save(outfile, writer=writer, dpi=dpi)
        return

    if ext == 'gif':
        print(f"Saving GIF {outfile}...")
        Writer = animation.writers['imagemagick']
        writer = Writer(fps=fps)
        ani.save(outfile, writer=writer, dpi=dpi)
        return

    raise ValueError(f"Don't know how to write '{ext}', sorry")

def plot_h_ext(h_ext):
    h_ext = np.array(h_ext)

    # (time, 2): a global vector signal
    # (time, H, W, 2): a local vector signal on a grid
    assert h_ext.ndim == 2 or h_ext.ndim == 4

    if h_ext.ndim == 2:
        # (time, 2): a global vector signal
        plt.plot(h_ext[:,0], label="x")
        plt.plot(h_ext[:,1], label="y")
        plt.xlabel('t')
        plt.ylabel('h_ext')
        plt.legend()

    elif h_ext.ndim == 4:
        # (time, H, W, 2): a local vector signal on a grid
        nrows = h_ext.shape[1]
        ncols = h_ext.shape[2]
        fig = plt.gcf()
        axes = fig.subplots(nrows, ncols, squeeze=False, sharex=True, sharey=True)
        for row, col in np.ndindex((nrows, ncols)):
            ax = axes[row, col]
            j = nrows - row - 1
            i = col
            ax.text(.95, .95, f"({j},{i})", horizontalalignment='right',
                    verticalalignment='top', transform=ax.transAxes,
                    zorder=0)
            ax.plot(h_ext[:,j,i,0], label="x")
            ax.plot(h_ext[:,j,i,1], label="y")
            if i == 0:
                ax.set_ylabel('h_ext')
            else:
                ax.tick_params('y', labelleft=False)
            if j == 0:
                ax.set_xlabel('t')
            else:
                ax.tick_params('x', labelbottom=False)
            if (j, i) == (nrows - 1, ncols - 1):
                ax.legend(loc='upper left', bbox_to_anchor=(1.0, 1.05))
        plt.subplots_adjust(wspace=0, hspace=0)

def gridless_crop(x, crop_percent):
    """ Return the indices of x that are inside the crop """
    minx, maxx = np.min(x[:,0]), np.max(x[:,0])
    miny, maxy = np.min(x[:,1]), np.max(x[:,1])
    crop_width = ((np.array(crop_percent)/100).T * [maxy - miny, maxx - minx]).T
    return np.logical_and(
                np.logical_and(
                    x[:,0] >= minx + crop_width[1][0],
                    x[:,0] <= maxx - crop_width[1][1]
                ),
                np.logical_and(
                    x[:,1] >= miny + crop_width[0][0],
                    x[:,1] <= maxy - crop_width[0][1]
                )
    )

def rotate_points(x, y, theta):
    return x * np.cos(theta) - y * np.sin(theta), x * np.sin(theta) + y * np.cos(theta)

def plot_astroid(b=1, c=1, beta=3, gamma=3, hc=1, rotation=0, resolution=361,
        angle_range=(0, 2*np.pi), ax=None, **kwargs):

    thetas = np.linspace(angle_range[0], angle_range[1], resolution)

    h_par = b * np.cos(thetas)
    h_perp = c * (1 - ((h_par/b)**2)**(1/gamma))**(beta/2)
    h_perp[thetas>np.pi] *= -1

    h_par, h_perp = rotate_points(h_par, h_perp, np.deg2rad(rotation))

    if ax is None:
        ax = plt.gca()

    ax.set_aspect('equal')

    return ax.plot(hc * h_par, hc * h_perp, **kwargs)

def plot_color_wheel(angles=8, angle_start=0, angle_range=360, radians=False,
        center_radius=0.13, ax=None, **kwargs):

    if np.isscalar(angles):
        angle_stop = angle_start + angle_range
        angles = np.arange(angle_start, angle_stop, (angle_stop - angle_start) / angles)

    if not radians:
        angles = np.radians(angles)

    ax = ax or plt.gca()

    circle = plt.Circle([0, 0], radius=center_radius, fc='white', zorder=100)
    ax.add_patch(circle)

    XY = np.zeros((len(angles), 2))
    UV = np.column_stack([np.cos(angles), np.sin(angles)])

    kwargs.setdefault('width', .1)

    return plot_vectors(XY, UV, pivot='tail', **kwargs)



import numpy as np
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal

from flatspin import SquareSpinIceClosed

def test_init(opencl):
    si = SquareSpinIceClosed(size=(5,4), **opencl)
    N = 5 * 5 + 4 * 6

    assert si.size == (5, 4)
    assert si.spin_count == N
    assert si.num_neighbors == 8

    si = SquareSpinIceClosed(size=(2,2), init="ground", **opencl) #check groundstate init
    assert_array_equal(si.spin, [ 1, -1, -1,  1, -1, -1,  1,  1, -1,  1,  1, -1])


def test_labels():
    si = SquareSpinIceClosed(size=(2,2))
    assert_array_equal(si.labels,
            [(0,0), (0,1),
             (1,0), (1,1), (1,2),
             (2,0), (2,1),
             (3,0), (3,1), (3,2),
             (4,0), (4,1)])

def test_indexof():
    si = SquareSpinIceClosed(size=(2,2))

    inds = [si.indexof(tuple(l)) for l in si.labels]
    assert_array_equal(inds, si.indices())

    # row 1
    assert_array_equal(si.L[1], [2, 3, 4])

    # col 1
    assert_array_equal(si.L[:,1], [1, 3, 6, 8, 11])

    # every other row
    assert_array_equal(si.L[::2], [0, 1, 5, 6, 10, 11])

    # every other column
    assert_array_equal(si.L[:,::2], [0, 2, 4, 5, 7, 9, 10])

def test_geometry():
    si = SquareSpinIceClosed(size=(3,2))
    L = si.L

    i_horiz = L[::2]
    i_vert = L[1::2]

    assert_array_equal(si.angle[i_horiz], 0)
    assert_allclose(si.angle[i_vert], np.pi/2)

    a = si.lattice_spacing
    assert_array_equal(si.pos[i_horiz, 0], a/2 + a*np.tile(np.arange(3), 3))
    assert_array_equal(si.pos[i_horiz, 1], a*np.repeat(np.arange(3), 3))

    assert_array_equal(si.pos[i_vert, 0], a*np.tile(np.arange(4), 2))
    assert_array_equal(si.pos[i_vert, 1], a/2 + a*np.repeat(np.arange(2), 4))

def test_num_neighbors():
    size = (10,10)

    # nearest neighborhood (default)
    si = SquareSpinIceClosed(size=size)
    assert si.num_neighbors == 8

    # global neighborhood
    si = SquareSpinIceClosed(size=size, neighbor_distance=np.inf)
    assert si.num_neighbors == si.spin_count - 1

def test_neighbors():
    size = (3,3)
    si = SquareSpinIceClosed(size=size)
    L = si.L

    # Horizontal lower left corner
    i = L[0,0]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[1,0], L[1,1], L[0,1], L[2,0]])

    # Horizontal lower center
    i = L[0,1]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[1,1], L[1,2], L[0,0], L[0,2], L[2,1]])

    # Horizontal lower right corner
    i = L[0,2]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[1,2], L[1,3], L[0,1], L[2,2]])

    # Vertical lower left corner
    i = L[1,0]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[0,0], L[2,0], L[3,0], L[1,1]])

    # Vertical left center
    i = L[1,1]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[0,0], L[0,1], L[2,0], L[2,1], L[3,1], L[1,0], L[1,2]])

    # Vertical lower right corner
    i = L[1,3]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[0,2], L[2,2], L[3,3], L[1,2]])

    # Horizontal left
    i = L[2,0]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[1,0], L[1,1], L[3,0], L[3,1], L[2,1], L[0,0], L[4,0]])

    # Horizontal center
    i = L[2,1]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[1,1], L[1,2], L[3,1], L[3,2], L[2,0], L[2,2], L[0,1], L[4,1]])

    # Horizontal right
    i = L[2,2]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[1,2], L[1,3], L[3,2], L[3,3], L[2,1], L[0,2], L[4,2]])

    # Vertical left
    i = L[3,0]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[2,0], L[4,0], L[1,0], L[5,0], L[3,1]])

    # Vertical left center
    i = L[3,1]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[2,0], L[2,1], L[4,0], L[4,1], L[1,1], L[5,1], L[3,0], L[3,2]])

    # Vertical right
    i = L[3,3]
    ns = list(si.neighbors(i))
    assert set(ns) == set([L[2,2], L[4,2], L[1,3], L[5,3], L[3,2]])

def test_spin_dipolar_field(opencl):
    # set lattice spacing such that NN islands have coupling strength 1.5
    a = 2*np.cos(np.pi/4)
    size = (4,4)
    si = SquareSpinIceClosed(size=size, lattice_spacing=a, **opencl)
    L = si.L

    j_NN = np.asarray([1.5, 0.5])
    j_L = np.asarray([1/np.sqrt(2), 0])
    j_T = np.asarray([1/(2*np.sqrt(2)), 0])

    assert si.spin_dipolar_field(L[2,1], L[3,1]) == approx( (-j_NN[0],  j_NN[1]))
    assert si.spin_dipolar_field(L[2,1], L[3,2]) == approx( j_NN)
    assert si.spin_dipolar_field(L[2,1], L[1,1]) == approx( j_NN)
    assert si.spin_dipolar_field(L[2,1], L[1,2]) == approx( (-j_NN[0],  j_NN[1]))

    assert si.spin_dipolar_field(L[2,1], L[2,2]) == approx( j_L )
    assert si.spin_dipolar_field(L[2,1], L[2,0]) == approx( j_L )

    assert si.spin_dipolar_field(L[2,1], L[0,1]) == approx( (-j_T[0], j_T[1]))
    assert si.spin_dipolar_field(L[2,1], L[4,1]) == approx( (-j_T[0], j_T[1]))

def test_dipolar_field(opencl):
    # set lattice spacing such that NN islands have coupling strength 1.5
    a = 2*np.cos(np.pi/4)
    size = (4,4)
    si = SquareSpinIceClosed(size=size, alpha=1.0, lattice_spacing=a, **opencl)
    L = si.L

    j_NN = np.asarray([1.5, 0.5])
    j_L = np.asarray([1/np.sqrt(2), 0])
    j_T = np.asarray([1/(2*np.sqrt(2)), 0])

    assert si.dipolar_field(L[2,1]) == approx(2*j_L - 2*j_T + [0, 4] * j_NN)
    si.flip(L[2,1])
    assert si.dipolar_field(L[2,1]) == approx(-2*j_L + 2*j_T - [0, 4] * j_NN)

    si.flip(L[3,1])
    assert si.dipolar_field(L[2,1]) == approx(-2*j_L + 2*j_T - [2, 2] * j_NN)
    si.flip(L[1,2])
    assert si.dipolar_field(L[2,1]) == approx(-2*j_L + 2*j_T - [4, 0] *j_NN)

def test_external_field(opencl):
    a = 2*np.cos(np.pi/4)
    size = (4,4)
    si = SquareSpinIceClosed(size=size, lattice_spacing=a, **opencl)
    L = si.L

    assert si.external_field(L[0,0]) == approx([0, 0])
    assert si.external_field(L[1,0]) == approx([0, 0])

    # Horizontal
    si.set_h_ext((1,0))
    assert si.external_field(L[0,0]) == approx([1, 0])
    assert si.external_field(L[1,0]) == approx([0,-1])

    # Vertical
    si.set_h_ext((0,1))
    assert si.external_field(L[0,0]) == approx([0, 1])
    assert si.external_field(L[1,0]) == approx([1, 0])

    # 45 degrees
    si.set_h_ext((1,1))
    assert si.external_field(L[0,0]) == approx([1, 1])
    assert si.external_field(L[1,0]) == approx([1,-1])

def test_total_field(opencl):
    si = SquareSpinIceClosed(size=(4,4), **opencl)
    si.randomize()

    i = si.L[2,2]
    assert tuple(si.total_field(i)) == tuple(si.dipolar_field(i) + si.external_field(i))

def test_dipolar_fields(opencl):
    si = SquareSpinIceClosed(size=(5,3), **opencl)
    si.randomize()

    h_dip = si.dipolar_fields()
    for i in si.indices():
        assert tuple(h_dip[i]) == tuple(si.dipolar_field(i))

def test_external_fields(opencl):
    si = SquareSpinIceClosed(size=(8,5), **opencl)
    si.randomize()

    # global external field
    si.set_h_ext((-1,2))

    h_ext = si.external_fields()
    for i in si.indices():
        assert tuple(h_ext[i]) == approx(tuple(si.external_field(i)))

    # local external field
    h_ext = np.random.uniform(-10, 10, (si.spin_count, 2))
    si.set_h_ext(h_ext)

    h_ext = si.external_fields()
    for i in si.indices():
        assert tuple(h_ext[i]) == approx(tuple(si.external_field(i)))

def test_total_fields(opencl):
    si = SquareSpinIceClosed(size=(4,4), **opencl)
    si.randomize()

    h_tot = si.total_fields()
    for i in si.indices():
        assert tuple(h_tot[i]) == tuple(si.total_field(i))


@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_flippable(opencl, switching):
    si = SquareSpinIceClosed(size=(4,4), hc=1.0, neighbor_distance=np.inf, switching=switching, **opencl)

    assert len(list(si.flippable())) == 0

    si.flip(si.L[0,0])

    E = si.switching_energy()

    flippable = [i for i in si.indices() if E[i] > 0]
    assert_array_equal(si.flippable(), flippable)

    si.polarize()
    si.set_h_ext([-10, -10])
    assert len(si.flippable()) == si.spin_count

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_alpha(opencl, switching):
    hc = 1.0
    h_ext = -(hc + 1.1)
    si = SquareSpinIceClosed(size=(4,4), alpha=1.0, hc=hc, neighbor_distance=np.inf, switching=switching, h_ext=(h_ext, 0), **opencl)
    h_dip = si.dipolar_fields()
    h_external = si.external_fields()
    h_tot = si.total_fields()
    n_flip = len(si.flippable())

    assert 0 < n_flip < si.spin_count

    # adjust alpha only
    alpha = 0.5
    si = SquareSpinIceClosed(size=(4,4), hc=hc, alpha=alpha, neighbor_distance=np.inf, switching=switching, h_ext=(h_ext, 0), **opencl)
    assert np.all(alpha * h_dip == si.dipolar_fields())
    assert np.all(h_external == si.external_fields())
    assert np.all(alpha * h_dip + h_external == si.total_fields())

    # adjust alpha, rescale hc and h_ext
    alpha = 0.5
    hc *= alpha
    h_ext *= alpha
    si = SquareSpinIceClosed(size=(4,4), hc=hc, alpha=alpha, neighbor_distance=np.inf, switching=switching, h_ext=(h_ext, 0), **opencl)

    assert np.all(alpha * h_dip == si.dipolar_fields())
    assert np.all(alpha * h_external == si.external_fields())
    assert np.all(alpha * h_tot == si.total_fields())
    assert len(si.flippable()) == n_flip

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_relax(opencl, switching):
    si = SquareSpinIceClosed(size=(4,4), alpha=1.0, switching=switching, neighbor_distance=np.inf, hc=1.0, **opencl)
    si.flip(si.L[1,1])
    assert si.relax() > 0
    assert len(si.flippable()) == 0

def test_energy(opencl):
    si = SquareSpinIceClosed(size=(4,4), hc=1.0, **opencl)
    E0 = si.energy()
    si.flip(si.L[0,0])
    E1 = si.energy()
    assert E1[0] > E0[0]

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_total_energy(opencl, switching):
    si = SquareSpinIceClosed(size=(6,5), switching=switching, neighbor_distance=np.inf, hc=1.0, **opencl)

    E = si.total_energy()
    si.flip(si.L[0,0])
    assert si.total_energy() > E

    si.randomize()
    E = si.total_energy()
    while si.step():
        assert si.total_energy() < E
        E = si.total_energy()

def test_find_vertices():
    si = SquareSpinIceClosed(size=(4,4))
    L = si.L
    vi, vj, indices = si.find_vertices()

    assert_array_equal(vi, [1, 1, 1, 3, 3, 3, 5, 5, 5])
    assert_array_equal(vj, [1, 3, 5, 1, 3, 5, 1, 3, 5])

    assert tuple(indices[0]) == (L[1,1], L[2,0], L[2,1], L[3,1])
    assert tuple(indices[1]) == (L[1,2], L[2,1], L[2,2], L[3,2])
    assert tuple(indices[2]) == (L[1,3], L[2,2], L[2,3], L[3,3])

    assert tuple(indices[3]) == (L[3,1], L[4,0], L[4,1], L[5,1])
    assert tuple(indices[4]) == (L[3,2], L[4,1], L[4,2], L[5,2])
    assert tuple(indices[5]) == (L[3,3], L[4,2], L[4,3], L[5,3])

    assert tuple(indices[6]) == (L[5,1], L[6,0], L[6,1], L[7,1])
    assert tuple(indices[7]) == (L[5,2], L[6,1], L[6,2], L[7,2])
    assert tuple(indices[8]) == (L[5,3], L[6,2], L[6,3], L[7,3])

def test_vertex_type():
    si = SquareSpinIceClosed(size=(4,4))
    spin = [1, 1, 1, 1, # horiz
            1, 1, -1, 1, 1, # vert
            -1, 1, 1, -1, # horiz
            1, -1, -1, 1, 1, # vert
            -1, 1, -1, -1, # horiz
            1, 1, 1, 1, 1, # vert
            1, 1, 1, -1, # horiz
            1, -1, 1, -1, 1, # vert
            1, 1, 1, 1] # horiz
    si.set_spin(spin)

    vertex = si.vertices()

    assert si.vertex_type(vertex[0]) == 1
    assert si.vertex_type(vertex[1]) == 2
    assert si.vertex_type(vertex[2]) == 3
    assert si.vertex_type(vertex[3]) == 4
    assert si.vertex_type(vertex[4]) == 1
    assert si.vertex_type(vertex[5]) == 2
    assert si.vertex_type(vertex[6]) == 3
    assert si.vertex_type(vertex[7]) == 2
    assert si.vertex_type(vertex[8]) == 4

def test_spin_grid(opencl):
    si = SquareSpinIceClosed(size=(2,1), **opencl)
    spin = [-1, 1,
            1, -1, 1,
            1, -1]
    si.set_spin(spin)

    # grid_spacing == lattice_spacing
    grid = si.spin_grid()
    assert grid.shape == (3, 5, 2)
    expect = [[[0,0], [-1,0], [0,0],  [1,0], [0,0]],
              [[0,1],  [0,0], [0,-1], [0,0], [0,1]],
              [[0,0],  [1,0], [0,0], [-1,0], [0,0]]]
    assert_allclose(grid, expect)

    # grid_spacing < lattice_spacing
    grid = si.spin_grid(si.lattice_spacing/4)
    assert grid.shape == (5, 9, 2)
    expect = [[[0,0], [0,0], [-1,0], [0,0], [0,0],  [0,0], [1,0],  [0,0], [0,0]],
              [[0,0], [0,0],  [0,0], [0,0], [0,0],  [0,0], [0,0],  [0,0], [0,0]],
              [[0,1], [0,0],  [0,0], [0,0], [0,-1], [0,0], [0,0],  [0,0], [0,1]],
              [[0,0], [0,0],  [0,0], [0,0], [0,0],  [0,0], [0,0],  [0,0], [0,0]],
              [[0,0], [0,0],  [1,0], [0,0], [0,0],  [0,0], [-1,0], [0,0], [0,0]]]
    assert_allclose(grid, expect)

    # grid_spacing > lattice_spacing
    grid = si.spin_grid(si.lattice_spacing)
    assert grid.shape == (2, 3, 2)
    expect = [[[-1,1], [1,-1], [0,1]],
              [[1,0],  [-1,0], [0,0]]]
    assert_allclose(grid, expect)

    # sum of all spins
    grid = si.spin_grid(si.spin_count * si.lattice_spacing)
    assert grid.shape == (1, 1, 2)
    expect = [[[0,1]]]
    assert_allclose(grid, expect)
